from metadrive import MetaDriveEnv


def test_config_consistency():
    env = MetaDriveEnv({"vehicle_config": {"lidar": {"num_lasers": 999}}})
    try:
        env.reset()
        assert env.vehicle.config["lidar"]["num_lasers"] == 999
    finally:
        env.close()


def test_config_consistency_2():
    # env = MetaDriveEnv({"map_config": {"config": "OO"}})
    the_config = 11
    env = MetaDriveEnv({"map_config": {"config": the_config}})
    try:
        env.reset()
        assert env.current_map.config["config"] == the_config
        assert all([v.config["config"] == the_config for v in env.maps.values()])
        assert env.config["map_config"]["config"] == the_config
    finally:
        env.close()

    the_config = 11
    env = MetaDriveEnv({"map": the_config})
    try:
        env.reset()
        assert env.current_map.config["config"] == the_config
        assert all([v.config["config"] == the_config for v in env.maps.values()])
        assert env.config["map_config"]["config"] == the_config
    finally:
        env.close()

    the_config = "OO"
    env = MetaDriveEnv({"map": the_config})
    try:
        env.reset()
        assert env.current_map.config["config"] == the_config
        assert all([v.config["config"] == the_config for v in env.maps.values()])
        assert env.config["map_config"]["config"] == the_config
    finally:
        env.close()

    the_config = "OO"
    env = MetaDriveEnv({"map_config": {"config": the_config, "type": "block_sequence"}})
    try:
        env.reset()
        assert env.current_map.config["config"] == the_config
        assert all([v.config["config"] == the_config for v in env.maps.values()])
        assert env.config["map_config"]["config"] == the_config
    finally:
        env.close()


if __name__ == '__main__':
    # test_config_consistency()
    test_config_consistency_2()



import copy

from metadrive.component.lane.point_lane import PointLane
from metadrive.utils.math import compute_angular_velocity
from metadrive.utils.math import norm


def get_max_valid_indicis(track, current_index):
    states = track["state"]
    assert states["valid"][current_index], "Current index should be valid"
    end = len(states["valid"])
    for i, valid in enumerate(states["valid"][current_index + 1:], current_index + 1):
        if not valid:
            end = i
            break
    return current_index, end


def get_idm_route(traj_points, width=2):
    traj = PointLane(traj_points, width)
    return traj


def parse_object_state(object_dict, time_idx, check_last_state=False, sim_time_interval=0.1):
    states = object_dict["state"]

    epi_length = len(states["position"])
    if time_idx < 0:
        time_idx = epi_length + time_idx

    if time_idx >= len(states["position"]):
        time_idx = len(states["position"]) - 1
    if check_last_state:
        for current_idx in range(time_idx):
            p_1 = states["position"][current_idx][:2]
            p_2 = states["position"][current_idx + 1][:2]
            if norm(p_1[0] - p_2[0], p_1[1] - p_2[1]) > 100:
                time_idx = current_idx
                break

    ret = {k: v[time_idx] for k, v in states.items()}

    ret["position"] = states["position"][time_idx, :2]
    ret["velocity"] = states["velocity"][time_idx]

    ret["heading_theta"] = states["heading"][time_idx]

    ret["heading"] = ret["heading_theta"]

    # optional keys with scalar value:
    for k in ["length", "width", "height"]:
        if k in states:
            ret[k] = float(states[k][time_idx])

    ret["valid"] = states["valid"][time_idx]
    if time_idx < len(states["position"]) - 1 and states["valid"][time_idx] and states["valid"][time_idx + 1]:
        angular_velocity = compute_angular_velocity(
            initial_heading=states["heading"][time_idx],
            final_heading=states["heading"][time_idx + 1],
            dt=sim_time_interval
        )
        ret["angular_velocity"] = angular_velocity
    else:
        ret["angular_velocity"] = 0

    # Retrieve vehicle type
    ret["vehicle_class"] = None
    if "spawn_info" in object_dict["metadata"]:
        type_module, type_cls_name = object_dict["metadata"]["spawn_info"]["type"]
        import importlib
        module = importlib.import_module(type_module)
        cls = getattr(module, type_cls_name)
        ret["vehicle_class"] = cls

    return ret


def parse_full_trajectory(object_dict):
    positions = object_dict["state"]["position"]
    index = len(positions)
    for current_idx in range(len(positions) - 1):
        p_1 = positions[current_idx][:2]
        p_2 = positions[current_idx + 1][:2]
        if norm(p_1[0] - p_2[0], p_1[1] - p_2[1]) > 100:
            index = current_idx
            break
    positions = positions[:index]
    trajectory = copy.deepcopy(positions[:, :2])

    return trajectory



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import Vec2, Vec3, TextNode, Vec4
from direct.gui.OnscreenText import OnscreenText

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.rpobject import RPObject


class Text(RPObject):
    """ Simple wrapper around OnscreenText, providing a simpler interface """
    def __init__(self, text="", parent=None, x=0, y=0, size=10, align="left", color=None, may_change=False, font=None):
        """ Constructs a new text. The parameters are almost equal to the
        parameters of OnscreenText """
        RPObject.__init__(self)

        if color is None:
            color = Vec3(1)

        align_mode = TextNode.A_left

        if align == "center":
            align_mode = TextNode.A_center
        elif align == "right":
            align_mode = TextNode.A_right

        if font is None:
            font = Globals.font
            # Should always have a global font. Never use the default panda font!
            assert font

        self._initial_pos = Vec2(x, -y)
        self._node = OnscreenText(
            text=text,
            parent=parent,
            pos=self._initial_pos,
            scale=size,
            align=align_mode,
            fg=Vec4(color.x, color.y, color.z, 1.0),
            font=font,
            mayChange=may_change
        )

    @property
    def node(self):
        """ Returns a handle to the internlally used node """
        return self._node

    def set_text(self, text):
        """ Changes the text, remember to pass may_change to the constructor,
        otherwise this method does not work. """
        self._node["text"] = text

    def get_initial_pos(self):
        """ Returns the initial position of the text. This can be used for
        animations """
        return self._initial_pos

    def pos_interval(self, *args, **kwargs):
        """ Returns a pos interval, this is a wrapper around
        NodePath.pos_interval """
        return self._node.pos_interval(*args, **kwargs)



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import bpy
import collections
import math

try:
    from collections.abc import Callable
except ImportError:
    from collections import Callable

from metadrive.third_party.kitsunetsuki.base.matrices import get_bone_matrix

from panda3d.core import CS_zup_right
from panda3d.egg import EggTable, EggXfmSAnim


class AnimationMixin(object):
    def make_action(self, node, armature, action):
        # <-- root table
        egg_root_table = EggTable('')

        # <-- animation bundle
        egg_animation = EggTable('Armature')
        egg_animation.set_table_type(EggTable.TT_bundle)

        # <-- skeleton table
        egg_skeleton = EggTable('<skeleton>')

        # setup bones
        egg_joints = {}
        egg_joints_anims = {}
        for bone_name, bone in armature.data.bones.items():
            # <-- joint table
            egg_joint = EggTable(bone_name)

            # <-- xfm_anim_s
            egg_xfm_anim_s = EggXfmSAnim('xform', CS_zup_right)
            # set order from glTF
            # i, j, k - scale -> s
            # h, p, r - rotation
            # x, y, z - location -> ?
            # egg_xfm_anim_s.set_order('shprxyz')
            egg_xfm_anim_s.set_fps(bpy.context.scene.render.fps / bpy.context.scene.render.fps_base)
            egg_joint.add_child(egg_xfm_anim_s)
            egg_joints_anims[bone_name] = egg_xfm_anim_s
            # xfm_anim_s -->

            if bone.parent:
                egg_joints[bone.parent.name].add_child(egg_joint)
            else:  # root bone
                egg_skeleton.add_child(egg_joint)

            egg_joints[bone_name] = egg_joint
            # joint table -->

        # set animation data
        frame_start = bpy.context.scene.frame_start
        frame_end = bpy.context.scene.frame_end
        if action:
            armature.animation_data.action = action
            frame_start, frame_end = action.frame_range

        frame = float(frame_start)
        frame_int = None
        while frame <= frame_end:
            # switch frame
            if frame_int != math.floor(frame):
                frame_int = math.floor(frame)
                bpy.context.scene.frame_current = frame_int
                bpy.context.scene.frame_set(frame_int)

            if isinstance(self._speed_scale, Callable):
                speed_scale = self._speed_scale(frame_int)
            else:
                speed_scale = self._speed_scale

            # switch subframe
            if speed_scale != 1:
                bpy.context.scene.frame_subframe = frame - frame_int

            # save bone matrices
            for bone_name, bone in armature.pose.bones.items():
                s_anim = egg_joints_anims[bone_name]
                bone_matrix = get_bone_matrix(bone, armature)

                # egg_joints_anims[bone_name].add_data(matrix_to_panda(bone_matrix))

                i, j, k = bone_matrix.to_scale()
                s_anim.add_component_data('i', i)
                s_anim.add_component_data('j', j)
                s_anim.add_component_data('k', k)

                # if bone.parent:
                #     p, r, h = tuple(map(math.degrees, bone_matrix.to_euler()))  # YABEE
                #     s_anim.add_component_data('h', h)
                #     s_anim.add_component_data('p', p)
                #     s_anim.add_component_data('r', r)
                # else:
                #     h, r, p = tuple(map(math.degrees, bone_matrix.to_euler('XZY')))
                #     s_anim.add_component_data('h', -h)
                #     s_anim.add_component_data('p', p)
                #     s_anim.add_component_data('r', r)

                # hpr == zxy ?
                p, r, h = tuple(map(math.degrees, bone_matrix.to_euler('YXZ')))
                s_anim.add_component_data('h', h)
                s_anim.add_component_data('p', p)
                s_anim.add_component_data('r', r)

                x, y, z = bone_matrix.to_translation()
                s_anim.add_component_data('x', x)
                s_anim.add_component_data('y', y)
                s_anim.add_component_data('z', z)

            # advance to the next frame
            frame += speed_scale

        egg_animation.add_child(egg_skeleton)
        # skeleton table -->

        egg_root_table.add_child(egg_animation)
        # animation bundle -->

        # root table -->

        node.add_child(egg_root_table)



import copy

from metadrive.envs.marl_envs.marl_inout_roundabout import MultiAgentRoundaboutEnv


class ChangeNEnv(MultiAgentRoundaboutEnv):
    def __init__(self, config=None):
        self._raw_input_config = copy.deepcopy(config)
        super(ChangeNEnv, self).__init__(config)

    def close_and_reset_num_agents(self, num_agents):
        config = copy.deepcopy(self._raw_input_config)
        self.close()
        config["num_agents"] = num_agents
        super(ChangeNEnv, self).__init__(config)


def test_change_agent_num():
    e = ChangeNEnv({})
    try:
        for num_agents in [1, 5, 10, 30]:
            e.close_and_reset_num_agents(num_agents)
            e.reset()
            for _ in range(100):
                e.step(e.action_space.sample())
                # e.render("topdown")
            # print("Hi!!! We are in environment now! Current agents: ", len(e.vehicles.keys()))
            e.close()
    finally:
        e.close()


if __name__ == '__main__':
    test_change_agent_num()



import logging
from collections import namedtuple

import math
import numpy as np
from panda3d.bullet import BulletGhostNode, BulletSphereShape
from panda3d.core import NodePath

from metadrive.constants import CamMask, CollisionGroup
from metadrive.engine.asset_loader import AssetLoader
from metadrive.engine.engine_utils import get_engine
from metadrive.utils.coordinates_shift import panda_vector
from metadrive.utils.math import panda_vector, get_laser_end

detect_result = namedtuple("detect_result", "cloud_points detected_objects")


def add_cloud_point_vis(
    point_x, point_y, height, num_lasers, laser_index, ANGLE_FACTOR, MARK_COLOR0, MARK_COLOR1, MARK_COLOR2
):
    f = laser_index / num_lasers if ANGLE_FACTOR else 1
    f *= 0.9
    f += 0.1
    return laser_index, (point_x, point_y, height), (f * MARK_COLOR0, f * MARK_COLOR1, f * MARK_COLOR2)


def perceive(
    cloud_points, detector_mask, mask, lidar_range, perceive_distance, heading_theta, vehicle_position_x,
    vehicle_position_y, num_lasers, height, physics_world, extra_filter_node, require_colors, ANGLE_FACTOR, MARK_COLOR0,
    MARK_COLOR1, MARK_COLOR2
):
    cloud_points.fill(1.0)
    detected_objects = []
    colors = []
    pg_start_position = panda_vector(vehicle_position_x, vehicle_position_y, height)

    for laser_index in range(num_lasers):
        if (detector_mask is not None) and (not detector_mask[laser_index]):
            # update vis
            if require_colors:
                point_x, point_y = get_laser_end(
                    lidar_range, perceive_distance, laser_index, heading_theta, vehicle_position_x, vehicle_position_y
                )
                point_x, point_y, point_z = panda_vector(point_x, point_y, height)
                colors.append(
                    add_cloud_point_vis(
                        point_x, point_y, height, num_lasers, laser_index, ANGLE_FACTOR, MARK_COLOR0, MARK_COLOR1,
                        MARK_COLOR2
                    )
                )
            continue

        # # coordinates problem here! take care
        point_x, point_y = get_laser_end(
            lidar_range, perceive_distance, laser_index, heading_theta, vehicle_position_x, vehicle_position_y
        )
        laser_end = panda_vector(point_x, point_y, height)
        result = physics_world.rayTestClosest(pg_start_position, laser_end, mask)
        node = result.getNode()
        if node in extra_filter_node:
            # Fall back to all tests.
            results = physics_world.rayTestAll(pg_start_position, laser_end, mask)
            hits = results.getHits()
            hits = sorted(hits, key=lambda ret: ret.getHitFraction())
            for result in hits:
                if result.getNode() in extra_filter_node:
                    continue
                detected_objects.append(result)
                cloud_points[laser_index] = result.getHitFraction()
                laser_end = result.getHitPos()
                break
        else:
            cloud_points[laser_index] = result.getHitFraction()
            if result.hasHit():
                laser_end = result.getHitPos()
            if node:
                detected_objects.append(result)
        if require_colors:
            colors.append(
                add_cloud_point_vis(
                    laser_end[0], laser_end[1], height, num_lasers, laser_index, ANGLE_FACTOR, MARK_COLOR0, MARK_COLOR1,
                    MARK_COLOR2
                )
            )
    return cloud_points, detected_objects, colors


class DistanceDetector:
    """
    It is a module like lidar, used to detect sidewalk/center line or other static things
    """
    Lidar_point_cloud_obs_dim = 240
    DEFAULT_HEIGHT = 0.2

    # for vis debug
    MARK_COLOR = (51 / 255, 221 / 255, 1)
    ANGLE_FACTOR = False

    def __init__(self, num_lasers: int = 16, distance: float = 50, enable_show=True):
        # properties
        self._node_path_list = []

        self.available = True if num_lasers > 0 and distance > 0 else False
        parent_node_np: NodePath = get_engine().render
        self.origin = parent_node_np.attachNewNode("Could_points")
        show = enable_show and (AssetLoader.loader is not None)
        self.dim = num_lasers
        self.num_lasers = num_lasers
        self.perceive_distance = distance
        self.height = self.DEFAULT_HEIGHT
        self.radian_unit = 2 * np.pi / num_lasers if self.num_lasers > 0 else None
        self.start_phase_offset = 0
        self._lidar_range = np.arange(0, self.num_lasers) * self.radian_unit + self.start_phase_offset

        # override these properties to decide which elements to detect and show
        self.origin.hide(CamMask.RgbCam | CamMask.Shadow | CamMask.Shadow | CamMask.DepthCam)
        self.mask = CollisionGroup.BrokenLaneLine
        self.cloud_points_vis = [] if show else None
        logging.debug("Load Vehicle Module: {}".format(self.__class__.__name__))
        if show:
            for laser_debug in range(self.num_lasers):
                ball = AssetLoader.loader.loadModel(AssetLoader.file_path("models", "box.bam"))
                ball.setScale(0.5)
                ball.setColor(0., 0.5, 0.5, 1)
                ball.reparentTo(self.origin)
                self.cloud_points_vis.append(ball)
            # self.origin.flattenStrong()

    def perceive(self, base_vehicle, physics_world, detector_mask: np.ndarray = None):
        assert self.available
        extra_filter_node = set(base_vehicle.dynamic_nodes)
        vehicle_position = base_vehicle.position
        heading_theta = base_vehicle.heading_theta
        assert not isinstance(detector_mask, str), "Please specify detector_mask either with None or a numpy array."
        cloud_points, detected_objects, colors = perceive(
            cloud_points=np.ones((self.num_lasers, ), dtype=float),
            detector_mask=detector_mask.astype(dtype=np.uint8) if detector_mask is not None else None,
            mask=self.mask,
            lidar_range=self._lidar_range,
            perceive_distance=self.perceive_distance,
            heading_theta=heading_theta,
            vehicle_position_x=vehicle_position[0],
            vehicle_position_y=vehicle_position[1],
            num_lasers=self.num_lasers,
            height=self.height,
            physics_world=physics_world,
            extra_filter_node=extra_filter_node if extra_filter_node else set(),
            require_colors=self.cloud_points_vis is not None,
            ANGLE_FACTOR=self.ANGLE_FACTOR,
            MARK_COLOR0=self.MARK_COLOR[0],
            MARK_COLOR1=self.MARK_COLOR[1],
            MARK_COLOR2=self.MARK_COLOR[2]
        )
        if self.cloud_points_vis is not None:
            for laser_index, pos, color in colors:
                self.cloud_points_vis[laser_index].setPos(pos)
                self.cloud_points_vis[laser_index].setColor(*color)
        return detect_result(cloud_points=cloud_points.tolist(), detected_objects=detected_objects)

    def _add_cloud_point_vis(self, laser_index, pos):
        self.cloud_points_vis[laser_index].setPos(pos)
        f = laser_index / self.num_lasers if self.ANGLE_FACTOR else 1
        self.cloud_points_vis[laser_index].setColor(
            f * self.MARK_COLOR[0], f * self.MARK_COLOR[1], f * self.MARK_COLOR[2]
        )

    def _get_laser_end(self, laser_index, heading_theta, vehicle_position):
        point_x = self.perceive_distance * math.cos(self._lidar_range[laser_index] + heading_theta) + \
                  vehicle_position[0]
        point_y = self.perceive_distance * math.sin(self._lidar_range[laser_index] + heading_theta) + \
                  vehicle_position[1]
        laser_end = panda_vector((point_x, point_y), self.height)
        return laser_end

    def destroy(self):
        if self.cloud_points_vis:
            for vis_laser in self.cloud_points_vis:
                vis_laser.removeNode()
        self.origin.removeNode()
        for np in self._node_path_list:
            np.detachNode()
            np.removeNode()

    def set_start_phase_offset(self, angle: float):
        """
        Change the start phase of lidar lasers
        :param angle: phasse offset in [degree]
        """
        self.start_phase_offset = np.deg2rad(angle)
        self._lidar_range = np.arange(0, self.num_lasers) * self.radian_unit + self.start_phase_offset

    def __del__(self):
        logging.debug("Lidar is destroyed.")

    def detach_from_world(self):
        if isinstance(self.origin, NodePath):
            self.origin.detachNode()

    def attach_to_world(self, engine):
        if isinstance(self.origin, NodePath):
            self.origin.reparentTo(engine.render)


class SideDetector(DistanceDetector):
    def __init__(self, num_lasers: int = 2, distance: float = 50, enable_show=True):
        super(SideDetector, self).__init__(num_lasers, distance, enable_show)
        self.set_start_phase_offset(90)
        self.origin.hide(CamMask.RgbCam | CamMask.Shadow | CamMask.Shadow | CamMask.DepthCam)
        self.mask = CollisionGroup.ContinuousLaneLine


class LaneLineDetector(SideDetector):
    MARK_COLOR = (1, 77 / 255, 77 / 255)

    def __init__(self, num_lasers: int = 2, distance: float = 50, enable_show=True):
        super(SideDetector, self).__init__(num_lasers, distance, enable_show)
        self.set_start_phase_offset(90)
        self.origin.hide(CamMask.RgbCam | CamMask.Shadow | CamMask.Shadow | CamMask.DepthCam)
        self.mask = CollisionGroup.ContinuousLaneLine | CollisionGroup.BrokenLaneLine



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import Mat4, PerspectiveLens, LVector3, LVecBase2i, LVecBase2f
from panda3d.core import BoundingSphere


class ShadowSource(object):
    """ Please refer to the native C++ implementation for docstrings and comments.
    This is just the python implementation, which does not contain documentation! """
    def __init__(self):
        self._slot = -1
        self._needs_update = True
        self._resolution = 512
        self._mvp = 0.0
        self._region = LVecBase2i(-1)
        self._region_uv = LVecBase2f(0.0)
        self._bounds = BoundingSphere()

    def set_resolution(self, resolution):
        self._resolution = resolution

    def set_needs_update(self, flag):
        self._needs_update = flag

    def set_slot(self, slot):
        self._slot = slot

    def set_region(self, region, region_uv):
        self._region = region
        self._region_uv = region_uv

    def clear_region(self):
        self._region = LVecBase2i(-1)
        self._region_uv = LVecBase2f(0.0)

    def get_bounds(self):
        return self._bounds

    def set_perspective_lens(self, fov, near_plane, far_plane, pos, direction):
        transform_mat = Mat4.translate_mat(-pos)
        temp_lens = PerspectiveLens(fov, fov)
        temp_lens.set_film_offset(0, 0)
        temp_lens.set_near_far(near_plane, far_plane)
        temp_lens.set_view_vector(direction, LVector3.up())
        self.set_matrix_lens(transform_mat * temp_lens.get_projection_mat())

        hexahedron = temp_lens.make_bounds()
        center = (hexahedron.get_min() + hexahedron.get_max()) * 0.5
        self._bounds = BoundingSphere(pos + center, (hexahedron.get_max() - center).length())

    def set_matrix_lens(self, mvp):
        self._mvp = mvp
        self.set_needs_update(True)

    def has_region(self):
        return (self._region.x >= 0 and self._region.y >= 0 and self._region.z >= 0 and self._region.w >= 0)

    def has_slot(self):
        return self._slot >= 0

    def get_slot(self):
        return self._slot

    def get_needs_update(self):
        return not self.has_region or self._needs_update

    def get_resolution(self):
        return self._resolution

    def get_mvp(self):
        return self._mvp

    def get_region(self):
        return self._region

    def get_uv_region(self):
        return self._region_uv

    def write_to_command(self, cmd):
        cmd.push_mat4(self._mvp)
        cmd.push_vec4(self._region_uv)



import matplotlib.pyplot as plt

from metadrive import MetaDriveEnv
from metadrive.component.map.city_map import CityMap
from metadrive.engine.engine_utils import initialize_engine, close_engine, set_global_random_seed
from metadrive.utils.draw_top_down_map import draw_top_down_map


def _t(num_blocks):
    default_config = MetaDriveEnv.default_config()
    initialize_engine(default_config)
    set_global_random_seed(0)
    try:
        map_config = default_config["map_config"]
        map_config.update(dict(type="block_num", config=num_blocks))
        map = CityMap(map_config)
        fig = draw_top_down_map(map, (1024, 1024))
        plt.imshow(fig, cmap="bone")
        plt.xticks([])
        plt.yticks([])
        plt.title("Building a City with {} blocks!".format(num_blocks))
        plt.show()
    finally:
        close_engine()


if __name__ == "__main__":
    _t(20)



from panda3d.core import NodePath, PGTop, TextNode, CardMaker, Vec3

from metadrive.constants import CamMask
from metadrive.engine.core.image_buffer import ImageBuffer


class VehiclePanel(ImageBuffer):
    PARA_VIS_LENGTH = 12
    PARA_VIS_HEIGHT = 1
    MAX_SPEED = 120
    BUFFER_W = 2
    BUFFER_H = 1
    CAM_MASK = CamMask.PARA_VIS
    GAP = 4.1
    TASK_NAME = "update panel"
    display_region_size = [2 / 3, 1, 0.8, 1.0]

    def __init__(self, engine, *, cuda):
        if engine.win is None:
            return
        self.aspect2d_np = NodePath(PGTop("aspect2d"))
        self.aspect2d_np.show(self.CAM_MASK)
        self.para_vis_np = {}

        tmp_node_path_list = []
        # make_buffer_func, make_camera_func = engine.win.makeTextureBuffer, engine.makeCamera

        # don't delete the space in word, it is used to set a proper position
        for i, np_name in enumerate(["Steering", " Throttle", "     Brake", "    Speed"]):
            text = TextNode(np_name)
            text.setText(np_name)
            text.setSlant(0.1)
            textNodePath = self.aspect2d_np.attachNewNode(text)
            tmp_node_path_list.append(textNodePath)

            textNodePath.setScale(0.052)
            text.setFrameColor(0, 0, 0, 1)
            text.setTextColor(0, 0, 0, 1)
            text.setFrameAsMargin(-self.GAP, self.PARA_VIS_LENGTH, 0, 0)
            text.setAlign(TextNode.ARight)
            textNodePath.setPos(-1.125111, 0, 0.9 - i * 0.08)
            if i != 0:
                cm = CardMaker(np_name)
                cm.setFrame(0, self.PARA_VIS_LENGTH - 0.21, -self.PARA_VIS_HEIGHT / 2 + 0.1, self.PARA_VIS_HEIGHT / 2)
                cm.setHasNormals(True)
                card = textNodePath.attachNewNode(cm.generate())
                tmp_node_path_list.append(card)

                card.setPos(0.21, 0, 0.22)
                self.para_vis_np[np_name.lstrip()] = card
            else:
                # left
                name = "Left"
                cm = CardMaker(name)
                cm.setFrame(
                    0, (self.PARA_VIS_LENGTH - 0.4) / 2, -self.PARA_VIS_HEIGHT / 2 + 0.1, self.PARA_VIS_HEIGHT / 2
                )
                cm.setHasNormals(True)
                card = textNodePath.attachNewNode(cm.generate())
                tmp_node_path_list.append(card)

                card.setPos(0.2 + self.PARA_VIS_LENGTH / 2, 0, 0.22)
                self.para_vis_np[name] = card
                # right
                name = "Right"
                cm = CardMaker(np_name)
                cm.setFrame(
                    -(self.PARA_VIS_LENGTH - 0.1) / 2, 0, -self.PARA_VIS_HEIGHT / 2 + 0.1, self.PARA_VIS_HEIGHT / 2
                )
                cm.setHasNormals(True)
                card = textNodePath.attachNewNode(cm.generate())
                tmp_node_path_list.append(card)

                card.setPos(0.2 + self.PARA_VIS_LENGTH / 2, 0, 0.22)
                self.para_vis_np[name] = card
        super(VehiclePanel, self).__init__(
            self.BUFFER_W,
            self.BUFFER_H,
            Vec3(-0.9, -1.01, 0.78),
            self.BKG_COLOR,
            parent_node=self.aspect2d_np,
            engine=engine
        )
        # self.add_display_region(self.display_region_size)
        self._node_path_list.extend(tmp_node_path_list)

    def update_vehicle_state(self, vehicle):
        steering, throttle_brake, speed = vehicle.steering, vehicle.throttle_brake, vehicle.speed_km_h
        if throttle_brake < 0:
            self.para_vis_np["Throttle"].setScale(0, 1, 1)
            self.para_vis_np["Brake"].setScale(-throttle_brake, 1, 1)
        elif throttle_brake > 0:
            self.para_vis_np["Throttle"].setScale(throttle_brake, 1, 1)
            self.para_vis_np["Brake"].setScale(0, 1, 1)
        else:
            self.para_vis_np["Throttle"].setScale(0, 1, 1)
            self.para_vis_np["Brake"].setScale(0, 1, 1)

        steering_value = abs(steering)
        if steering < 0:
            self.para_vis_np["Left"].setScale(steering_value, 1, 1)
            self.para_vis_np["Right"].setScale(0, 1, 1)
        elif steering > 0:
            self.para_vis_np["Right"].setScale(steering_value, 1, 1)
            self.para_vis_np["Left"].setScale(0, 1, 1)
        else:
            self.para_vis_np["Right"].setScale(0, 1, 1)
            self.para_vis_np["Left"].setScale(0, 1, 1)
        speed_value = speed / self.MAX_SPEED
        self.para_vis_np["Speed"].setScale(speed_value, 1, 1)

    def remove_display_region(self):
        self.buffer.set_active(False)
        super(VehiclePanel, self).remove_display_region()

    def add_display_region(self, display_region):
        super(VehiclePanel, self).add_display_region(display_region)
        self.buffer.set_active(True)
        self.origin.reparentTo(self.aspect2d_np)

    def destroy(self):
        super(VehiclePanel, self).destroy()
        for para in self.para_vis_np.values():
            para.removeNode()
        self.aspect2d_np.removeNode()

    def track(self, base_object):
        # for compatibility
        pass



import logging
import math
import os
from collections import deque
from typing import Union, Optional

import numpy as np
import seaborn as sns
from panda3d._rplight import RPSpotLight
from panda3d.bullet import BulletVehicle, BulletBoxShape, ZUp
from panda3d.core import Material, Vec3, TransformState
from panda3d.core import NodePath

from metadrive.base_class.base_object import BaseObject
from metadrive.component.lane.abs_lane import AbstractLane
from metadrive.component.lane.circular_lane import CircularLane
from metadrive.component.lane.point_lane import PointLane
from metadrive.component.lane.straight_lane import StraightLane
from metadrive.component.pg_space import VehicleParameterSpace, ParameterSpace
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.component.sensors.distance_detector import SideDetector, LaneLineDetector
from metadrive.component.sensors.lidar import Lidar
from metadrive.component.vehicle_navigation_module.edge_network_navigation import EdgeNetworkNavigation
from metadrive.component.vehicle_navigation_module.node_network_navigation import NodeNetworkNavigation
from metadrive.constants import MetaDriveType, CollisionGroup
from metadrive.engine.asset_loader import AssetLoader
from metadrive.engine.engine_utils import get_engine, engine_initialized
from metadrive.engine.physics_node import BaseRigidBodyNode
from metadrive.utils import Config, safe_clip_for_small_array
from metadrive.utils.math import get_vertical_vector, norm, clip
from metadrive.utils.math import wrap_to_pi
from metadrive.utils.pg.utils import rect_region_detection
from metadrive.utils.utils import get_object_from_node


class BaseVehicleState:
    def __init__(self):
        self.init_state_info()

    def init_state_info(self):
        """
        Call this before reset()/step()
        """
        self.crash_vehicle = False
        self.crash_human = False
        self.crash_object = False
        self.crash_sidewalk = False
        self.crash_building = False

        # traffic light
        self.red_light = False
        self.yellow_light = False
        self.green_light = False

        # lane line detection
        self.on_yellow_continuous_line = False
        self.on_white_continuous_line = False
        self.on_broken_line = False

        # contact results, a set containing objects type name for rendering
        self.contact_results = set()


class BaseVehicle(BaseObject, BaseVehicleState):
    """
    Vehicle chassis and its wheels index
                    0       1
                    II-----II
                        |
                        |  <---chassis/wheelbase
                        |
                    II-----II
                    2       3
    """
    COLLISION_MASK = CollisionGroup.Vehicle
    PARAMETER_SPACE = ParameterSpace(VehicleParameterSpace.BASE_VEHICLE)
    MAX_LENGTH = 10
    MAX_WIDTH = 2.5
    MAX_STEERING = 60

    # LENGTH = None
    # WIDTH = None
    # HEIGHT = None

    TIRE_RADIUS = None
    TIRE_MODEL_CORRECT = 1  # correct model left right error
    TIRE_TWO_SIDED = False  # tires for some vehicles need two-sided enabled for correctly visualization
    LATERAL_TIRE_TO_CENTER = None
    TIRE_WIDTH = 0.4
    FRONT_WHEELBASE = None
    REAR_WHEELBASE = None
    LIGHT_POSITION = (-0.67, 1.62, 0.05)

    # MASS = None

    CHASSIS_TO_WHEEL_AXIS = 0.2
    SUSPENSION_LENGTH = 15
    SUSPENSION_STIFFNESS = 40

    # for random color choosing
    MATERIAL_COLOR_COEFF = 2  # to resist other factors, since other setting may make color dark
    MATERIAL_METAL_COEFF = 0.1  # 0-1
    MATERIAL_ROUGHNESS = 0.8  # smaller to make it more smooth, and reflect more light
    MATERIAL_SHININESS = 128  # 0-128 smaller to make it more smooth, and reflect more light
    MATERIAL_SPECULAR_COLOR = (3, 3, 3, 3)

    # control
    STEERING_INCREMENT = 0.05

    # save memory, load model once
    model_collection = {}
    path = None

    def __init__(
        self,
        vehicle_config: Union[dict, Config] = None,
        name: str = None,
        random_seed=None,
        position=None,
        heading=None
    ):
        """
        This Vehicle Config is different from self.get_config(), and it is used to define which modules to use, and
        module parameters. And self.physics_config defines the physics feature of vehicles, such as length/width
        :param vehicle_config: mostly, vehicle module config
        :param random_seed: int
        """
        # check
        assert vehicle_config is not None, "Please specify the vehicle config."
        assert engine_initialized(), "Please make sure game engine is successfully initialized!"

        # NOTE: it is the game engine, not vehicle drivetrain
        # self.engine = get_engine()
        BaseObject.__init__(self, name, random_seed, self.engine.global_config["vehicle_config"])
        BaseVehicleState.__init__(self)
        self.update_config(vehicle_config)
        use_special_color = self.config["use_special_color"]

        # build vehicle physics model
        vehicle_chassis = self._create_vehicle_chassis()
        self.add_body(vehicle_chassis.getChassis())
        self.system = vehicle_chassis
        self.chassis = self.origin
        self.wheels = self._create_wheel()

        # light experimental!
        self.light = None
        self._light_direction_queue = None
        self._light_models = None
        self.light_name = None

        # powertrain config
        self.increment_steering = self.config["increment_steering"]
        self.enable_reverse = self.config["enable_reverse"]
        self.max_steering = self.config["max_steering"]

        # visualization
        self._use_special_color = use_special_color
        self._add_visualization()

        # modules, get observation by using these modules
        self.navigation: Optional[NodeNetworkNavigation] = None
        self.lidar: Optional[Lidar] = None  # detect surrounding vehicles
        self.side_detector: Optional[SideDetector] = None  # detect road side
        self.lane_line_detector: Optional[LaneLineDetector] = None  # detect nearest lane lines

        # state info
        self.throttle_brake = 0.0
        self.steering = 0
        self.last_current_action = deque([(0.0, 0.0), (0.0, 0.0)], maxlen=2)
        self.last_position = (0, 0)
        self.last_heading_dir = self.heading
        self.dist_to_left_side = None
        self.dist_to_right_side = None
        self.last_velocity = 0
        self.last_speed = 0

        # step info
        self.out_of_route = None
        self.on_lane = None
        self.spawn_place = (0, 0)
        self._init_step_info()

        # others
        self._add_modules_for_vehicle()
        self.takeover = False
        self.expert_takeover = False
        self.energy_consumption = 0
        self.break_down = False

        # overtake_stat
        self.front_vehicles = set()
        self.back_vehicles = set()

        # if self.engine.current_map is not None:
        self.reset(position=position, heading=heading)

    def _add_modules_for_vehicle(self, ):
        """
        This function is related to the self.update_config, which will create modules if needed for resetting a new
        vehicle
        """
        config = self.config

        # add routing module
        self.add_navigation()  # default added

        # add distance detector/lidar
        self.side_detector = SideDetector(
            config["side_detector"]["num_lasers"], config["side_detector"]["distance"],
            self.engine.global_config["vehicle_config"]["show_side_detector"]
        )

        self.lane_line_detector = LaneLineDetector(
            config["lane_line_detector"]["num_lasers"], config["lane_line_detector"]["distance"],
            self.engine.global_config["vehicle_config"]["show_lane_line_detector"]
        )

        self.lidar = Lidar(
            config["lidar"]["num_lasers"], config["lidar"]["distance"],
            self.engine.global_config["vehicle_config"]["show_lidar"]
        )

    def _add_modules_for_vehicle_when_reset(self):
        config = self.config

        # add routing module
        if self.navigation is None:
            self.add_navigation()  # default added

        # add distance detector/lidar
        if self.side_detector is None:
            self.side_detector = SideDetector(
                config["side_detector"]["num_lasers"], config["side_detector"]["distance"],
                self.engine.global_config["vehicle_config"]["show_side_detector"]
            )

        if self.lane_line_detector is None:
            self.lane_line_detector = LaneLineDetector(
                config["lane_line_detector"]["num_lasers"], config["lane_line_detector"]["distance"],
                self.engine.global_config["vehicle_config"]["show_lane_line_detector"]
            )

        if self.lidar is None:
            self.lidar = Lidar(
                config["lidar"]["num_lasers"], config["lidar"]["distance"],
                self.engine.global_config["vehicle_config"]["show_lidar"]
            )

    def _init_step_info(self):
        # done info will be initialized every frame
        self.init_state_info()
        self.out_of_route = False  # re-route is required if is false
        self.on_lane = True  # on lane surface or not

    @staticmethod
    def _preprocess_action(action):
        if action is None:
            return None, {"raw_action": None}
        action = safe_clip_for_small_array(action, -1, 1)
        return action, {'raw_action': (action[0], action[1])}

    def before_step(self, action=None):
        """
        Save info and make decision before action
        """
        # init step info to store info before each step
        # if action is None:
        #     action = [0, 0]

        self._init_step_info()
        action, step_info = self._preprocess_action(action)

        self.last_position = self.position  # 2D vector
        self.last_velocity = self.velocity  # 2D vector
        self.last_speed = self.speed  # Scalar
        self.last_heading_dir = self.heading
        if action is not None:
            self.last_current_action.append(action)  # the real step of physics world is implemented in taskMgr.step()
        if self.increment_steering:
            self._set_incremental_action(action)
        else:
            self._set_action(action)
        return step_info

    def after_step(self):
        if self.navigation is not None:
            self.navigation.update_localization(self)
        self._state_check()
        self.update_dist_to_left_right()
        step_energy, episode_energy = self._update_energy_consumption()
        self.out_of_route = self._out_of_route()
        step_info = self._update_overtake_stat()
        my_policy = self.engine.get_policy(self.name)
        step_info.update(
            {
                "velocity": float(self.speed),
                "steering": float(self.steering),
                "acceleration": float(self.throttle_brake),
                "step_energy": step_energy,
                "episode_energy": episode_energy,
                "policy": my_policy.name if my_policy is not None else my_policy
            }
        )
        return step_info

    def _out_of_route(self):
        left, right = self._dist_to_route_left_right()
        return True if right < 0 or left < 0 else False

    def _update_energy_consumption(self):
        """
        The calculation method is from
        https://www.researchgate.net/publication/262182035_Reduction_of_Fuel_Consumption_and_Exhaust_Pollutant_Using_Intelligent_Transport_System
        default: 3rd gear, try to use ae^bx to fit it, dp: (90, 8), (130, 12)
        :return: None
        """
        distance = norm(self.last_position[0] - self.position[0], self.last_position[1] - self.position[1]) / 1000  # km
        step_energy = 3.25 * math.pow(np.e, 0.01 * self.speed_km_h) * distance / 100
        # step_energy is in Liter, we return mL
        step_energy = step_energy * 1000
        self.energy_consumption += step_energy  # L/100 km
        return step_energy, self.energy_consumption

    def reset(
        self,
        vehicle_config=None,
        name=None,
        random_seed=None,
        position: np.ndarray = None,
        heading: float = 0.0,
        *args,
        **kwargs
    ):
        """
        pos is a 2-d array, and heading is a float (unit degree)
        if pos is not None, vehicle will be reset to the position
        else, vehicle will be reset to spawn place
        """
        if name is not None:
            self.rename(name)

        if random_seed is not None:
            assert isinstance(random_seed, int)
            self.seed(random_seed)
            self.sample_parameters()
        if vehicle_config is not None:
            self.update_config(vehicle_config)

        # Update some modules that might not be initialized before
        self._add_modules_for_vehicle_when_reset()

        self.set_pitch(0)
        self.set_roll(0)
        if position is not None:
            # Highest priority
            pass
        elif self.config["spawn_position_heading"] is None:
            # spawn_lane_index has second priority
            map = self.engine.current_map
            assert map, "Map should not be None"
            lane = map.road_network.get_lane(self.config["spawn_lane_index"])
            position = lane.position(self.config["spawn_longitude"], self.config["spawn_lateral"])
            heading = lane.heading_theta_at(self.config["spawn_longitude"])
        else:
            assert self.config["spawn_position_heading"] is not None, "At least setting one initialization method"
            position = self.config["spawn_position_heading"][0]
            heading = self.config["spawn_position_heading"][1]

        self.spawn_place = position
        self.set_heading_theta(heading)
        self.set_static(False)
        # self.set_wheel_friction(self.config["wheel_friction"])

        if len(position) == 2:
            self.set_position(position, height=self.HEIGHT / 2)
        elif len(position) == 3:
            self.set_position(position[:2], height=position[-1])
        else:
            raise ValueError()

        self.reset_navigation()
        self.body.clearForces()
        self.body.setLinearVelocity(Vec3(0, 0, 0))
        self.body.setAngularVelocity(Vec3(0, 0, 0))
        self.system.resetSuspension()
        self._apply_throttle_brake(0.0)
        # np.testing.assert_almost_equal(self.position, pos, decimal=4)

        # done info
        self._init_step_info()

        # other info
        self.throttle_brake = 0.0
        self.steering = 0
        self.last_current_action = deque([(0.0, 0.0), (0.0, 0.0)], maxlen=2)
        self.last_position = self.spawn_place
        self.last_heading_dir = self.heading
        self.last_velocity = self.velocity  # 2D vector
        self.last_speed = self.speed  # Scalar

        self.update_dist_to_left_right()
        self.takeover = False
        self.energy_consumption = 0

        # overtake_stat
        self.front_vehicles = set()
        self.back_vehicles = set()
        self.expert_takeover = False
        if self.config["need_navigation"]:
            assert self.navigation

        if self.config["spawn_velocity"] is not None:
            self.set_velocity(self.config["spawn_velocity"], in_local_frame=self.config["spawn_velocity_car_frame"])

        # clean lights
        if self.config["light"]:
            self.add_light()
        else:
            self.remove_light()

        # self.add_light()

    """------------------------------------------- act -------------------------------------------------"""

    def remove_light(self):
        if self.light is not None:
            if self.use_render_pipeline:
                self.engine.render_pipeline.remove_light(self.light)
                self.engine.taskMgr.remove(self.light_name)
            self.light_name = None
            self.light = None
            for m in self._light_models:
                m.removeNode()
            self._light_models = None
            self._light_direction_queue = None

    def add_light(self):
        """
        Experimental feature
        """
        # assert self.use_render_pipeline, "Can be Enabled when using render pipeline"
        if self.light is None:
            self._light_models = []
            for y in [-1, 1]:
                light_model = self.loader.loadModel(AssetLoader.file_path("models", "sphere.egg"))
                light_model.reparentTo(self.origin)
                light_model.setPos(self.LIGHT_POSITION[0] * y, self.LIGHT_POSITION[1], self.LIGHT_POSITION[2])
                light_model.setScale(0.13, 0.05, 0.13)
                material = Material()
                material.setBaseColor((1, 1, 1, 1))
                material.setShininess(128)
                material.setEmission((1, 1, 1, 1))
                light_model.setMaterial(material, True)
                self._light_models.append(light_model)
            if self.use_render_pipeline:
                self.light_name = "light_{}".format(self.id)
                self.light = RPSpotLight()
                self.light.set_color_from_temperature(3 * 1000.0)
                self.light.setRadius(500)
                self.light.setFov(100)
                self.light.energy = 600
                self.light.casts_shadows = False
                self.light.shadow_map_resolution = 128
                self.engine.render_pipeline.add_light(self.light)
                self.engine.taskMgr.add(self._update_light_pos, self.light_name)
                self._light_direction_queue = []

    def _update_light_pos(self, task):
        pos = self.convert_to_world_coordinates([self.LENGTH / 2, 0], self.position)
        self.light.set_pos(*pos, self.get_z())
        self._light_direction_queue.append([*self.heading, 0])
        idx = min(len(self._light_direction_queue), 50)
        pos = np.mean(self._light_direction_queue[-idx:], axis=0)
        self.light.set_direction(pos[0], pos[1], pos[2])
        return task.cont

    def set_steering(self, steering):
        steering = float(steering)
        self.system.setSteeringValue(steering, 0)
        self.system.setSteeringValue(steering, 1)
        self.steering = steering

    def set_throttle_brake(self, throttle_brake):
        throttle_brake = float(throttle_brake)
        self._apply_throttle_brake(throttle_brake)
        self.throttle_brake = throttle_brake

    def _set_action(self, action):
        if action is None:
            return
        steering = action[0]
        self.throttle_brake = action[1]
        self.steering = steering
        self.system.setSteeringValue(self.steering * self.max_steering, 0)
        self.system.setSteeringValue(self.steering * self.max_steering, 1)
        self._apply_throttle_brake(action[1])

    def _set_incremental_action(self, action: np.ndarray):
        if action is None:
            return
        self.throttle_brake = action[1]
        self.steering += action[0] * self.STEERING_INCREMENT
        self.steering = clip(self.steering, -1, 1)
        steering = self.steering * self.max_steering
        self.system.setSteeringValue(steering, 0)
        self.system.setSteeringValue(steering, 1)
        self._apply_throttle_brake(action[1])

    def _apply_throttle_brake(self, throttle_brake):
        max_engine_force = self.config["max_engine_force"]
        max_brake_force = self.config["max_brake_force"]
        for wheel_index in range(4):
            if throttle_brake >= 0:
                self.system.setBrake(2.0, wheel_index)
                if self.speed_km_h > self.max_speed_km_h:
                    self.system.applyEngineForce(0.0, wheel_index)
                else:
                    self.system.applyEngineForce(max_engine_force * throttle_brake, wheel_index)
            else:
                if self.enable_reverse:
                    self.system.applyEngineForce(max_engine_force * throttle_brake, wheel_index)
                    self.system.setBrake(0, wheel_index)
                else:
                    self.system.applyEngineForce(0.0, wheel_index)
                    self.system.setBrake(abs(throttle_brake) * max_brake_force, wheel_index)

    """---------------------------------------- vehicle info ----------------------------------------------"""

    def update_dist_to_left_right(self):
        self.dist_to_left_side, self.dist_to_right_side = self._dist_to_route_left_right()

    def _dist_to_route_left_right(self):
        # TODO
        if self.navigation is None or self.navigation.current_ref_lanes is None:
            return 0, 0
        current_reference_lane = self.navigation.current_ref_lanes[0]
        _, lateral_to_reference = current_reference_lane.local_coordinates(self.position)
        lateral_to_left = lateral_to_reference + self.navigation.get_current_lane_width() / 2
        lateral_to_right = self.navigation.get_current_lateral_range(self.position, self.engine) - lateral_to_left
        return lateral_to_left, lateral_to_right

    # @property
    # def heading_theta(self):
    #     """
    #     Get the heading theta of vehicle, unit [rad]
    #     :return:  heading in rad
    #     """
    #     return wrap_to_pi(self.origin.getH() / 180 * math.pi)

    # @property
    # def velocity(self) -> np.ndarray:
    #     return self.speed * self.velocity_direction
    #
    # @property
    # def velocity_km_h(self) -> np.ndarray:
    #     return self.speed * self.velocity_direction * 3.6

    @property
    def chassis_velocity_direction(self):
        raise DeprecationWarning(
            "This API returns the direction of velocity which is approximately heading direction. "
            "Deprecate it and make things easy"
        )
        direction = self.system.getForwardVector()
        return np.asarray([direction[0], direction[1]])

    """---------------------------------------- some math tool ----------------------------------------------"""

    def heading_diff(self, target_lane):
        lateral = None
        if isinstance(target_lane, StraightLane):
            lateral = np.asarray(get_vertical_vector(target_lane.end - target_lane.start)[1])
        elif isinstance(target_lane, CircularLane):
            if not target_lane.is_clockwise():
                lateral = self.position - target_lane.center
            else:
                lateral = target_lane.center - self.position
        elif isinstance(target_lane, PointLane):
            lateral = target_lane.lateral_direction(target_lane.local_coordinates(self.position)[0])

        lateral_norm = norm(lateral[0], lateral[1])
        forward_direction = self.heading
        # print(f"Old forward direction: {self.forward_direction}, new heading {self.heading}")
        forward_direction_norm = norm(forward_direction[0], forward_direction[1])
        if not lateral_norm * forward_direction_norm:
            return 0
        cos = (
            (forward_direction[0] * lateral[0] + forward_direction[1] * lateral[1]) /
            (lateral_norm * forward_direction_norm)
        )
        # return cos
        # Normalize to 0, 1
        return clip(cos, -1.0, 1.0) / 2 + 0.5

    def lane_distance_to(self, vehicle, lane: AbstractLane = None) -> float:
        assert self.navigation is not None, "a routing and localization module should be added " \
                                            "to interact with other vehicles"
        if not vehicle:
            return np.nan
        if not lane:
            lane = self.lane
        return lane.local_coordinates(vehicle.position)[0] - lane.local_coordinates(self.position)[0]

    """-------------------------------------- for vehicle making ------------------------------------------"""

    @property
    def LENGTH(self):
        raise NotImplementedError()

    @property
    def HEIGHT(self):
        raise NotImplementedError()

    @property
    def WIDTH(self):
        raise NotImplementedError()

    def _create_vehicle_chassis(self):
        # self.LENGTH = type(self).LENGTH
        # self.WIDTH = type(self).WIDTH
        # self.HEIGHT = type(self).HEIGHT

        # assert self.LENGTH < BaseVehicle.MAX_LENGTH, "Vehicle is too large!"
        # assert self.WIDTH < BaseVehicle.MAX_WIDTH, "Vehicle is too large!"

        chassis = BaseRigidBodyNode(self.name, MetaDriveType.VEHICLE)
        self._node_path_list.append(chassis)

        chassis_shape = BulletBoxShape(Vec3(self.WIDTH / 2, self.LENGTH / 2, self.HEIGHT / 2))
        ts = TransformState.makePos(Vec3(0, 0, self.HEIGHT / 2))
        chassis.addShape(chassis_shape, ts)
        chassis.setDeactivationEnabled(False)
        chassis.notifyCollisions(True)  # advance collision check, do callback in pg_collision_callback

        physics_world = get_engine().physics_world
        vehicle_chassis = BulletVehicle(physics_world.dynamic_world, chassis)
        vehicle_chassis.setCoordinateSystem(ZUp)
        self.dynamic_nodes.append(vehicle_chassis)
        return vehicle_chassis

    def _add_visualization(self):
        if self.render:
            [path, scale, offset, HPR] = self.path
            if path not in BaseVehicle.model_collection:
                car_model = self.loader.loadModel(AssetLoader.file_path("models", path))
                car_model.setTwoSided(False)
                BaseVehicle.model_collection[path] = car_model
                car_model.setScale(scale)
                # model default, face to y
                car_model.setHpr(*HPR)
                car_model.setPos(offset[0], offset[1], offset[-1])
                car_model.setZ(-self.TIRE_RADIUS - self.CHASSIS_TO_WHEEL_AXIS + offset[-1])
            else:
                car_model = BaseVehicle.model_collection[path]
            car_model.instanceTo(self.origin)
            if self.config["random_color"]:
                material = Material()
                material.setBaseColor(
                    (
                        self.panda_color[0] * self.MATERIAL_COLOR_COEFF,
                        self.panda_color[1] * self.MATERIAL_COLOR_COEFF,
                        self.panda_color[2] * self.MATERIAL_COLOR_COEFF, 0.2
                    )
                )
                material.setMetallic(self.MATERIAL_METAL_COEFF)
                material.setSpecular(self.MATERIAL_SPECULAR_COLOR)
                material.setRefractiveIndex(1.5)
                material.setRoughness(self.MATERIAL_ROUGHNESS)
                material.setShininess(self.MATERIAL_SHININESS)
                material.setTwoside(False)
                self.origin.setMaterial(material, True)

    def _create_wheel(self):
        f_l = self.FRONT_WHEELBASE
        r_l = -self.REAR_WHEELBASE
        lateral = self.LATERAL_TIRE_TO_CENTER
        axis_height = self.TIRE_RADIUS - self.CHASSIS_TO_WHEEL_AXIS
        radius = self.TIRE_RADIUS
        wheels = []
        for k, pos in enumerate([Vec3(lateral, f_l, axis_height), Vec3(-lateral, f_l, axis_height),
                                 Vec3(lateral, r_l, axis_height), Vec3(-lateral, r_l, axis_height)]):
            wheel = self._add_wheel(pos, radius, True if k < 2 else False, True if k == 0 or k == 2 else False)
            wheels.append(wheel)
        return wheels

    def _add_wheel(self, pos: Vec3, radius: float, front: bool, left):
        wheel_np = self.origin.attachNewNode("wheel")
        self._node_path_list.append(wheel_np)

        if self.render:
            model = 'right_tire_front.gltf' if front else 'right_tire_back.gltf'
            model_path = AssetLoader.file_path("models", os.path.dirname(self.path[0]), model)
            wheel_model = self.loader.loadModel(model_path)
            wheel_model.setTwoSided(self.TIRE_TWO_SIDED)
            wheel_model.reparentTo(wheel_np)
            wheel_model.set_scale(1 * self.TIRE_MODEL_CORRECT if left else -1 * self.TIRE_MODEL_CORRECT)
        wheel = self.system.create_wheel()
        wheel.setNode(wheel_np.node())
        wheel.setChassisConnectionPointCs(pos)
        wheel.setFrontWheel(front)
        wheel.setWheelDirectionCs(Vec3(0, 0, -1))
        wheel.setWheelAxleCs(Vec3(1, 0, 0))

        wheel.setWheelRadius(radius)
        wheel.setMaxSuspensionTravelCm(self.SUSPENSION_LENGTH)
        wheel.setSuspensionStiffness(self.SUSPENSION_STIFFNESS)
        wheel.setWheelsDampingRelaxation(4.8)
        wheel.setWheelsDampingCompression(1.2)
        wheel_friction = self.config["wheel_friction"] if not self.config["no_wheel_friction"] else 0
        wheel.setFrictionSlip(wheel_friction)
        wheel.setRollInfluence(0.5)
        return wheel

    def add_navigation(self):
        if not self.config["need_navigation"]:
            return
        navi = self.config["navigation_module"]
        if navi is None:
            navi = NodeNetworkNavigation if self.engine.current_map.road_network_type == NodeRoadNetwork \
                else EdgeNetworkNavigation
        self.navigation = navi(
            # self.engine,
            show_navi_mark=self.engine.global_config["vehicle_config"]["show_navi_mark"],
            random_navi_mark_color=self.engine.global_config["vehicle_config"]["random_navi_mark_color"],
            show_dest_mark=self.engine.global_config["vehicle_config"]["show_dest_mark"],
            show_line_to_dest=self.engine.global_config["vehicle_config"]["show_line_to_dest"],
            panda_color=self.panda_color,
            name=self.name,
            vehicle_config=self.config
        )

    def reset_navigation(self):
        """
        Update map information that are used by this vehicle, after reset()
        This function will query the map about the spawn position and destination of current vehicle,
        and update the navigation module by feeding the information of spawn point and destination.

        For the spawn position, if it is not specify in the config["spawn_lane_index"], we will automatically
        select one lane based on the localization results.
        """
        if self.navigation is not None and self.config["need_navigation"]:
            self.navigation.reset(self)
            self.navigation.update_localization(self)

    def _state_check(self):
        """
        Check States and filter to update info
        """
        result_1 = self.engine.physics_world.static_world.contactTest(self.chassis.node(), True)
        result_2 = self.engine.physics_world.dynamic_world.contactTest(self.chassis.node(), True)
        contacts = set()
        for contact in result_1.getContacts() + result_2.getContacts():
            node0 = contact.getNode0()
            node1 = contact.getNode1()
            node = node0 if node1.getName() == MetaDriveType.VEHICLE else node1
            name = node.getName()
            if name == MetaDriveType.LINE_SOLID_SINGLE_WHITE:
                self.on_white_continuous_line = True
            elif name == MetaDriveType.LINE_SOLID_SINGLE_YELLOW:
                self.on_yellow_continuous_line = True
            elif name == MetaDriveType.LINE_BROKEN_SINGLE_YELLOW or name == MetaDriveType.LINE_BROKEN_SINGLE_WHITE:
                self.on_broken_line = True
            elif name == MetaDriveType.TRAFFIC_LIGHT:
                light = get_object_from_node(node)
                if light.status == MetaDriveType.LIGHT_GREEN:
                    self.green_light = True
                elif light.status == MetaDriveType.LIGHT_RED:
                    self.red_light = True
                elif light.status == MetaDriveType.LIGHT_YELLOW:
                    self.yellow_light = True
                elif light.status == MetaDriveType.LIGHT_UNKNOWN:
                    # unknown didn't add
                    continue
                else:
                    raise ValueError("Unknown light status: {}".format(light.status))
                name = light.status
            # they work with the function in collision_callback.py to double-check the collision
            elif name == MetaDriveType.VEHICLE:
                self.crash_vehicle = True
            elif name == MetaDriveType.BUILDING:
                self.crash_building = True
            elif MetaDriveType.is_traffic_object(name):
                self.crash_object = True
            elif name in [MetaDriveType.PEDESTRIAN, MetaDriveType.CYCLIST]:
                self.crash_human = True
            else:
                # didn't add
                continue
            contacts.add(name)
        # side walk detect
        res = rect_region_detection(
            self.engine,
            self.position,
            np.rad2deg(self.heading_theta),
            self.LENGTH,
            self.WIDTH,
            CollisionGroup.Sidewalk,
            in_static_world=True if not self.render else False
        )
        if res.hasHit() and res.getNode().getName() == MetaDriveType.BOUNDARY_LINE:
            self.crash_sidewalk = True
            contacts.add(MetaDriveType.BOUNDARY_LINE)

        # only for visualization detection
        if self.render:
            res = rect_region_detection(
                self.engine,
                self.position,
                np.rad2deg(self.heading_theta),
                self.LENGTH,
                self.WIDTH,
                CollisionGroup.LaneSurface,
                in_static_world=True if not self.engine.global_config["debug_static_world"] else False
            )
            if not res.hasHit() or res.getNode().getName() != MetaDriveType.LANE_SURFACE_STREET:
                contacts.add(MetaDriveType.GROUND)

        self.contact_results.update(contacts)

    def destroy(self):
        super(BaseVehicle, self).destroy()
        if self.navigation is not None:
            self.navigation.destroy()
        self.navigation = None
        self.wheels = None
        if self.side_detector is not None:
            self.side_detector.destroy()
            self.side_detector = None
        if self.lane_line_detector is not None:
            self.lane_line_detector.destroy()
            self.lane_line_detector = None
        if self.lidar is not None:
            self.lidar.destroy()
            self.lidar = None
        if self.light is not None:
            self.remove_light()

    def set_velocity(self, direction, *args, **kwargs):
        super(BaseVehicle, self).set_velocity(direction, *args, **kwargs)
        self.last_velocity = self.velocity
        self.last_speed = self.speed

    def set_state(self, state):
        super(BaseVehicle, self).set_state(state)
        self.set_throttle_brake(float(state["throttle_brake"]))
        self.set_steering(float(state["steering"]))
        self.last_velocity = self.velocity
        self.last_speed = self.speed
        self.last_position = self.position
        self.last_heading_dir = self.heading

    def set_panda_pos(self, pos):
        super(BaseVehicle, self).set_panda_pos(pos)
        self.last_position = self.position

    def set_position(self, position, height=None):
        super(BaseVehicle, self).set_position(position, height)
        self.last_position = self.position

    def get_state(self):
        """
        Fetch more information
        """
        state = super(BaseVehicle, self).get_state()
        state.update(
            {
                "steering": self.steering,
                "throttle_brake": self.throttle_brake,
                "crash_vehicle": self.crash_vehicle,
                "crash_object": self.crash_object,
                "crash_building": self.crash_building,
                "crash_sidewalk": self.crash_sidewalk,
                "size": (self.LENGTH, self.WIDTH, self.HEIGHT),
                "length": self.LENGTH,
                "width": self.WIDTH,
                "height": self.HEIGHT,
            }
        )
        if self.navigation is not None:
            state.update(self.navigation.get_state())
        return state

    # def get_raw_state(self):
    #     ret = dict(position=self.position, heading=self.heading, velocity=self.velocity)
    #     return ret

    def get_dynamics_parameters(self):
        # These two can be changed on the fly
        max_engine_force = self.config["max_engine_force"]
        max_brake_force = self.config["max_brake_force"]

        # These two can only be changed in init
        wheel_friction = self.config["wheel_friction"]
        assert self.max_steering == self.config["max_steering"]
        max_steering = self.max_steering

        mass = self.config["mass"] if self.config["mass"] else self.MASS

        ret = dict(
            max_engine_force=max_engine_force,
            max_brake_force=max_brake_force,
            wheel_friction=wheel_friction,
            max_steering=max_steering,
            mass=mass
        )
        return ret

    def _update_overtake_stat(self):
        if self.config["overtake_stat"] and self.lidar.available:
            surrounding_vs = self.lidar.get_surrounding_vehicles()
            routing = self.navigation
            ckpt_idx = routing._target_checkpoints_index
            for surrounding_v in surrounding_vs:
                if surrounding_v.lane_index[:-1] == (routing.checkpoints[ckpt_idx[0]], routing.checkpoints[ckpt_idx[1]
                                                                                                           ]):
                    if self.lane.local_coordinates(self.position)[0] - \
                            self.lane.local_coordinates(surrounding_v.position)[0] < 0:
                        self.front_vehicles.add(surrounding_v)
                        if surrounding_v in self.back_vehicles:
                            self.back_vehicles.remove(surrounding_v)
                    else:
                        self.back_vehicles.add(surrounding_v)
        return {"overtake_vehicle_num": self.get_overtake_num()}

    def get_overtake_num(self):
        return len(self.front_vehicles.intersection(self.back_vehicles))

    def __del__(self):
        super(BaseVehicle, self).__del__()
        # self.engine = None
        self.lidar = None
        self.navigation = None
        self.wheels = None

    @property
    def reference_lanes(self):
        return self.navigation.current_ref_lanes

    def set_wheel_friction(self, new_friction):
        raise ValueError()
        for wheel in self.wheels:
            wheel.setFrictionSlip(new_friction)

    @property
    def overspeed(self):
        return True if self.lane.speed_limit < self.speed_km_h else False

    @property
    def replay_done(self):
        return self._replay_done if hasattr(self, "_replay_done") else (
            self.crash_building or self.crash_vehicle or
            # self.on_white_continuous_line or
            self.on_yellow_continuous_line
        )

    @property
    def current_action(self):
        return self.last_current_action[-1]

    @property
    def last_action(self):
        return self.last_current_action[0]

    def detach_from_world(self, physics_world):
        if self.navigation is not None:
            self.navigation.detach_from_world()
        if self.lidar is not None:
            self.lidar.detach_from_world()
        if self.side_detector is not None:
            self.side_detector.detach_from_world()
        if self.lane_line_detector is not None:
            self.lane_line_detector.detach_from_world()
        super(BaseVehicle, self).detach_from_world(physics_world)

    def attach_to_world(self, parent_node_path, physics_world):
        if self.config["show_navi_mark"] and self.config["need_navigation"]:
            self.navigation.attach_to_world(self.engine)
        if self.lidar is not None and self.config["show_lidar"]:
            self.lidar.attach_to_world(self.engine)
        if self.side_detector is not None and self.config["show_side_detector"]:
            self.side_detector.attach_to_world(self.engine)
        if self.lane_line_detector is not None and self.config["show_lane_line_detector"]:
            self.lane_line_detector.attach_to_world(self.engine)
        super(BaseVehicle, self).attach_to_world(parent_node_path, physics_world)

    def set_break_down(self, break_down=True):
        self.break_down = break_down
        # self.set_static(True)

    @property
    def max_speed_km_h(self):
        return self.config["max_speed_km_h"]

    @property
    def max_speed_m_s(self):
        return self.config["max_speed_km_h"] / 3.6

    @property
    def top_down_length(self):
        return self.LENGTH

    @property
    def top_down_width(self):
        return self.WIDTH

    @property
    def lane(self):
        return self.navigation.current_lane

    @property
    def lane_index(self):
        return self.navigation.current_lane.index

    @property
    def panda_color(self):
        c = super(BaseVehicle, self).panda_color
        if self._use_special_color:
            color = sns.color_palette("colorblind")
            rand_c = color[2]  # A pretty green
            c = rand_c
        return c

    def before_reset(self):
        for obj in [self.navigation, self.lidar, self.side_detector, self.lane_line_detector]:
            if obj is not None and hasattr(obj, "before_reset"):
                obj.before_reset()

    """------------------------------------------- overwrite -------------------------------------------------"""

    def convert_to_world_coordinates(self, vector, origin):
        return super(BaseVehicle, self).convert_to_world_coordinates([-vector[-1], vector[0]], origin)

    def convert_to_local_coordinates(self, vector, origin):
        ret = super(BaseVehicle, self).convert_to_local_coordinates(vector, origin)
        return np.array([ret[1], -ret[0]])

    @property
    def heading_theta(self):
        return wrap_to_pi(super(BaseVehicle, self).heading_theta + np.pi / 2)

    def set_heading_theta(self, heading_theta, in_rad=True) -> None:
        """
        Set heading theta for this object. Vehicle local frame has a 90 degree offset
        :param heading_theta: float in rad
        :param in_rad: when set to True, heading theta should be in rad, otherwise, in degree
        """
        super(BaseVehicle, self).set_heading_theta(heading_theta - np.pi / 2, in_rad)
        self.last_heading_dir = self.heading

    @property
    def roll(self):
        """
        Return the roll of this object
        """
        return np.deg2rad(self.origin.getR())

    def set_roll(self, roll):
        self.origin.setR(roll)

    @property
    def pitch(self):
        """
        Return the pitch of this object
        """
        return np.deg2rad(self.origin.getP())

    def set_pitch(self, pitch):
        self.origin.setP(pitch)

    def show_coordinates(self):
        if self.coordinates_debug_np is not None:
            self.coordinates_debug_np.reparentTo(self.origin)
            return
        height = self.HEIGHT + 0.2
        self.coordinates_debug_np = NodePath("debug coordinate")
        # 90 degrees offset
        x = self.engine.draw_line_3d([0, 0, height], [0, 2, height], [1, 1, 1, 1], 2)
        y = self.engine.draw_line_3d([0, 0, height], [-1, 0, height], [1, 1, 1, 1], 2)
        z = self.engine.draw_line_3d([0, 0, height], [0, 0, height + 0.5], [1, 1, 1, 1], 2)
        x.reparentTo(self.coordinates_debug_np)
        y.reparentTo(self.coordinates_debug_np)
        z.reparentTo(self.coordinates_debug_np)
        self.coordinates_debug_np.reparentTo(self.origin)



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

# This tool offers an interface to configure the time of day settings

from __future__ import print_function, division

import os
import sys
import time
from threading import Thread

# Change to the current directory
os.chdir(os.path.join(os.path.dirname(os.path.realpath(__file__))))

# Add the render pipeline to the path
sys.path.insert(0, "../../")

from rplibs.six import iteritems  # noqa
from rplibs.pyqt_imports import *  # noqa

from curve_widget import CurveWidget  # noqa

from rpcore.pluginbase.manager import PluginManager  # noqa
from rpcore.mount_manager import MountManager  # noqa
from rpcore.util.network_communication import NetworkCommunication  # noqa

from ui.main_window_generated import Ui_MainWindow  # noqa
from ui.point_insert_dialog_generated import Ui_Dialog as Ui_PointDialog  # noqa


class PointDialog(QDialog, Ui_PointDialog):
    def __init__(self, parent):
        QDialog.__init__(self, parent)
        self.setupUi(self)
        qt_connect(self.btn_insert, "clicked()", lambda: self.done(1))

    def get_value(self):
        time = self.ipt_time.time()
        val = self.ipt_value.value()
        return time, val


class DayTimeEditor(QMainWindow, Ui_MainWindow):
    """ This is the main editor class which handles the user interface """
    def __init__(self):

        # Init mounts
        self._mount_mgr = MountManager(None)
        self._mount_mgr.mount()

        self._plugin_mgr = PluginManager(None)
        self._plugin_mgr.load()

        QMainWindow.__init__(self)
        self.setupUi()
        self._tree_widgets = []
        self._cmd_queue = set()

        self._selected_setting_handle = None
        self._selected_setting = None
        self._selected_plugin = None
        self._current_time = 0.5

        self._update_settings_list()
        self._on_time_changed(self.time_slider.value())

        self._bg_thread = Thread(target=self.updateThread)
        self._bg_thread.start()

    def set_settings_visible(self, visibility):
        if not visibility:
            self.frame_current_setting.hide()
        else:
            self.frame_current_setting.show()

    def closeEvent(self, event):  # noqa
        event.accept()
        import os
        os._exit(1)

    def updateThread(self):  # noqa
        """ Seperate update thread """

        while True:
            if self._cmd_queue:
                cmd = self._cmd_queue.pop()
                if cmd == "settime":
                    NetworkCommunication.send_async(
                        NetworkCommunication.DAYTIME_PORT, "settime " + str(self._current_time)
                    )
                    continue
                elif cmd == "write_settings":
                    self._plugin_mgr.save_daytime_overrides("/$$rpconfig/daytime.yaml")
                    NetworkCommunication.send_async(NetworkCommunication.DAYTIME_PORT, "loadconf")
                else:
                    print("Unkown cmd:", cmd)

            time.sleep(0.1)

    def setupUi(self):  # noqa
        """ Setups the UI Components """
        Ui_MainWindow.setupUi(self, self)
        self.settings_tree.setColumnWidth(0, 160)
        self.settings_tree.expandAll()

        self.edit_widget = CurveWidget(self)
        self.edit_widget.set_change_handler(self._on_curve_edited)
        self.prefab_edit_widget.addWidget(self.edit_widget)

        qt_connect(self.time_slider, "valueChanged(int)", self._on_time_changed)
        qt_connect(self.settings_tree, "itemSelectionChanged()", self._on_setting_selected)
        qt_connect(self.btn_insert_point, "clicked()", self._insert_point)
        qt_connect(self.btn_reset, "clicked()", self._reset_settings)

    def _reset_settings(self):
        """ Resets the current plugins settings """
        # QMessageBox.warning(self, "Houston, we have a problem!",
        #     "This functionality is not yet implemented! Blame tobspr if you need it.\n\n"
        #     "On a more serious note, you can still hand-edit config/daytime.yaml.",
        #     QMessageBox.Ok, QMessageBox.Ok)

        # Ask the user if he's really sure about it
        msg = "Are you sure you want to reset the control points of '" +\
              self._selected_setting_handle.label + "'?\n"
        msg += "!! This cannot be undone !! They will be lost forever (a long time!)."
        reply = QMessageBox.question(self, "Warning", msg, QMessageBox.Yes, QMessageBox.No)
        if reply == QMessageBox.Yes:

            QMessageBox.information(self, "Success", "Control points have been reset!")
            default = self._selected_setting_handle.default
            self._selected_setting_handle.curves[0].set_single_value(default)
            self._update_settings_list()
            self._cmd_queue.add("write_settings")

    def _insert_point(self):
        """ Asks the user to insert a new point """
        dialog = PointDialog(self)
        if dialog.exec_():
            time, val = dialog.get_value()
            minutes = (time.hour() * 60 + time.minute()) / (24 * 60)

            if (val < self._selected_setting_handle.minvalue or val > self._selected_setting_handle.maxvalue):
                QMessageBox.information(self, "Invalid Value", "Value is out of setting range!", QMessageBox.Ok)
                return

            val_linear = self._selected_setting_handle.get_linear_value(val)
            self._selected_setting_handle.curves[0].append_cv(minutes, val_linear)
            self._cmd_queue.add("write_settings")

    def _update_tree_widgets(self):
        """ Updates the tree widgets """
        for setting_handle, widget in self._tree_widgets:
            value = setting_handle.get_scaled_value_at(self._current_time)
            formatted = setting_handle.format(value)
            widget.setText(1, formatted)

            if setting_handle.type == "color":
                widget.setBackground(1, QBrush(QColor(*value)))

    def _on_curve_edited(self):
        """ Called when the curve got edited in the curve widget """
        self._cmd_queue.add("write_settings")
        self._update_tree_widgets()

    def _on_setting_selected(self):
        """ Called when a setting got selected in the settings tree """
        selected = self.settings_tree.selectedItems()
        if len(selected) != 1:
            self._selected_setting = None
            self._selected_plugin = None
            self._selected_setting_handle = None
            self.edit_widget.set_curves([])
            self.set_settings_visible(False)
        else:
            selected = selected[0]

            self._selected_plugin = selected._plugin_id
            self._selected_setting = selected._setting_id
            self._selected_setting_handle = selected._setting_handle

            self.lbl_current_setting.setText(self._selected_setting_handle.label)
            self.lbl_setting_desc.setText(self._selected_setting_handle.description)

            self.edit_widget.set_curves(self._selected_setting_handle.curves)

            if self._selected_setting_handle.type == "color":
                self.edit_widget.set_unit_processor(lambda x: str(int(x * 255)))
                self.btn_insert_point.hide()
            else:
                self.edit_widget.set_unit_processor(
                    lambda x: self._selected_setting_handle.format(self._selected_setting_handle.get_scaled_value(x))
                )
                self.btn_insert_point.show()

            self.set_settings_visible(True)
            self._update_tree_widgets()

    def _on_time_changed(self, val):
        """ Handler when the time slider got moved """
        hour = val // (60 * 60 * 60)
        minute = (val // (60 * 60)) % 60
        ftime = float(val) / (24 * 60 * 60 * 60)

        self.time_label.setText(str(hour).zfill(2) + ":" + str(minute).zfill(2))
        self.time_float_label.setText("{:1.4f}".format(ftime))
        self.edit_widget.set_current_time(ftime)
        self._current_time = ftime
        self._update_tree_widgets()
        self._cmd_queue.add("settime")

    def _update_settings_list(self):
        """ Updates the list of visible settings """

        self.settings_tree.clear()
        self._tree_widgets = []

        first_item = None

        for plugin_id, plugin in iteritems(self._plugin_mgr.instances):

            daytime_settings = self._plugin_mgr.day_settings[plugin_id]

            if not daytime_settings:
                # Skip plugins with empty settings
                continue

            plugin_head = QTreeWidgetItem(self.settings_tree)
            plugin_head.setText(0, plugin.name)
            plugin_head.setFlags(Qt.ItemIsEnabled)
            font = QFont()
            font.setBold(True)
            if not self._plugin_mgr.is_plugin_enabled(plugin_id):
                plugin_head.setText(0, plugin.name)
            plugin_head.setFont(0, font)

            # Display all settings
            for setting, setting_handle in iteritems(daytime_settings):
                setting_item = QTreeWidgetItem(plugin_head)
                setting_item.setText(0, setting_handle.label)
                if PYQT_VERSION == 4:
                    setting_item.setTextColor(0, QColor(150, 150, 150))
                else:
                    setting_item.setForeground(0, QColor(150, 150, 150))
                setting_item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
                setting_item._setting_id = setting
                setting_item._setting_handle = setting_handle
                setting_item._plugin_id = plugin_id
                setting_item.setToolTip(0, setting_handle.description)
                setting_item.setToolTip(1, setting_handle.description)
                self._tree_widgets.append((setting_handle, setting_item))
                if not first_item:
                    first_item = setting_item

        self.settings_tree.expandAll()
        if first_item:
            self.settings_tree.setCurrentItem(first_item)


# Start application
app = QApplication(sys.argv)
qt_register_fonts()
editor = DayTimeEditor()
editor.show()
app.exec_()



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import bpy
import os

from metadrive.third_party.kitsunetsuki.base.collections import get_object_collection
from metadrive.third_party.kitsunetsuki.base.objects import (
    get_object_properties, is_collision, is_object_visible, set_active_object
)

from .geom import GeomMixin
from .material import MaterialMixin
from .texture import TextureMixin
from .vertex import VertexMixin

NOT_MERGED_TYPES = (
    'Portal',
    'Text',
    'Sprite',
    'Transparent',
    'Protected',
    'Dynamic',
    'Flipbook',
    'Slider',
    'Alpha',
)


class Exporter(GeomMixin, MaterialMixin, TextureMixin, VertexMixin):
    def __init__(self, args):
        self._inputs = args.inputs
        self._output = args.output

        if self._inputs:
            bpy.ops.wm.open_mainfile(filepath=self._inputs[0])
            for i in self._inputs[1:]:
                bpy.ops.wm.append(filepath=i)

        # export type
        self._export_type = args.export or 'scene'
        self._action = args.action  # animation/action name to export

        # render type
        self._render_type = args.render or 'default'

        # animations
        self._speed_scale = args.speed or 1

        # geom scale
        self._geom_scale = args.scale or 1

        # scripting
        self._script_names = (args.exec or '').split(',')
        self._script_locals = {}

        # merging
        self._merge = args.merge
        self._keep = args.keep

        # materials, textures, UVs
        self._no_materials = args.no_materials is True
        self._no_extra_uv = args.no_extra_uv is True
        self._no_textures = args.no_textures is True
        self._empty_textures = args.empty_textures
        self._set_origin = args.set_origin is True

    def get_cwd(self):
        if self._inputs:
            return os.path.dirname(self._inputs[0])
        else:
            return ''

    def execute_script(self, name):
        script = bpy.data.texts.get(name)
        if script:
            code = compile(script.as_string(), name, 'exec')
            exec(code, None, self._script_locals)

            if 'SPEED_SCALE' in self._script_locals:
                self._speed_scale = self._script_locals['SPEED_SCALE']

    def can_merge(self, obj):
        if not self._merge:
            return False

        collection = get_object_collection(obj)
        if not collection:
            return False

        if is_collision(obj):
            return False

        if not is_object_visible(obj):
            return False

        obj_props = get_object_properties(obj)
        if obj_props.get('type') in NOT_MERGED_TYPES:
            return False

        if obj.type == 'MESH':
            for material in obj.data.materials:
                if material.node_tree:
                    for node in material.node_tree.nodes:
                        if node.type == 'ATTRIBUTE':
                            return False

            return True

        return False

    def make_root_node(self):
        raise NotImplementedError()

    def make_empty(self, parent_node, obj):
        raise NotImplementedError()

    def make_mesh(self, parent_node, obj):
        raise NotImplementedError()

    def make_light(self, parent_node, obj):
        raise NotImplementedError()

    def make_armature(self, parent_node, obj):
        raise NotImplementedError()

    def make_animation(self, parent_node, obj=None):
        for child in bpy.data.objects:
            if not is_object_visible(child):
                continue

            if child.type == 'ARMATURE':
                if self._action:
                    action = bpy.data.actions[self._action]
                    self.make_action(parent_node, child, action)
                else:
                    for action_name, action in bpy.data.actions.items():
                        self.make_action(parent_node, child, action)

    def make_node(self, parent_node, obj=None):
        node = None

        if obj is None:
            node = parent_node

        else:
            if obj.type == 'EMPTY':
                node = self.make_empty(parent_node, obj)

            elif obj.type == 'ARMATURE':
                node = self.make_armature(parent_node, obj)

            elif obj.type == 'MESH':
                node = self.make_mesh(parent_node, obj)

            elif obj.type in ('LIGHT', 'LAMP'):
                if obj.data.type in ('SPOT', 'POINT'):
                    node = self.make_light(parent_node, obj)

        if node is None:
            return

        # make children of the current node
        if obj is None:  # root objects
            children = filter(lambda o: not o.parent, bpy.data.objects)
        else:  # children on current object
            children = obj.children

        for child in children:
            if not is_object_visible(child) and not is_collision(child):
                continue

            if self._export_type == 'collision':
                if child.type in ('ARMATURE', 'LIGHT', 'LAMP'):
                    continue

                if child.type == 'MESH' and not is_collision(child):
                    continue

            self.make_node(node, child)

    def convert(self):
        if self._script_names:
            for script_name in self._script_names:
                if script_name:
                    self.execute_script(script_name)

        if self._merge:
            for collection in bpy.data.collections:
                if collection.name == 'RigidBodyWorld':
                    continue

                objects = list(filter(self.can_merge, collection.objects))
                if not objects:
                    continue

                bpy.ops.object.select_all(action='DESELECT')
                for obj in objects:
                    obj.select_set(state=True)
                set_active_object(objects[0])

                context = {
                    'active_object': objects[0],
                    'selected_objects': objects,
                    'selected_editable_objects': objects,
                }

                bpy.ops.object.join(context)
                bpy.ops.object.select_all(action='DESELECT')
                bpy.context.view_layer.objects.active.name = collection.name
                set_active_object(None)

        self._root = self.make_root_node()

        if self._export_type == 'animation':
            self.make_animation(self._root)
        else:
            self.make_node(self._root)
            if self._export_type == 'all':
                self.make_animation(self._root)

        return self._root



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""
from __future__ import division

from panda3d.core import Vec3, Camera, OrthographicLens, PTAMat4, SamplerState

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.render_stage import RenderStage
from metadrive.render_pipeline.rpcore.util.generic import snap_shadow_map


class PSSMSceneShadowStage(RenderStage):
    """ This stage creates the shadow map which covers the whole important part
    of the scene. This is required because the shadow cascades only cover the
    view frustum, but many plugins (VXGI, EnvMaps) require a shadow map. """

    required_inputs = []
    required_pipes = []

    def __init__(self, pipeline):
        RenderStage.__init__(self, pipeline)
        self.resolution = 2048
        self.sun_vector = Vec3(0, 0, 1)
        self.sun_distance = 10.0
        self.pta_mvp = PTAMat4.empty_array(1)
        self.focus = None

        # Store last focus entirely for the purpose of being able to see
        # it in the debugger
        self.last_focus = None

    @property
    def produced_inputs(self):
        return {"PSSMSceneSunShadowMVP": self.pta_mvp}

    @property
    def produced_pipes(self):
        return {"PSSMSceneSunShadowMapPCF": (self.target.depth_tex, self.make_pcf_state())}

    def make_pcf_state(self):
        state = SamplerState()
        state.set_minfilter(SamplerState.FT_shadow)
        state.set_magfilter(SamplerState.FT_shadow)
        return state

    def request_focus(self, focus_point, focus_size):
        self.focus = (focus_point, focus_size)
        self.last_focus = self.focus

    @property
    def mvp(self):
        return Globals.base.render.get_transform(self.cam_node).get_mat() * \
            self.cam_lens.get_projection_mat()

    def update(self):
        if self._pipeline.task_scheduler.is_scheduled("pssm_scene_shadows"):
            if self.focus is None:
                # When no focus is set, there is no point in rendering the shadow map
                self.target.active = False
            else:
                focus_point, focus_size = self.focus

                self.cam_lens.set_near_far(0.0, 2 * (focus_size + self.sun_distance))
                self.cam_lens.set_film_size(2 * focus_size, 2 * focus_size)
                self.cam_node.set_pos(focus_point + self.sun_vector * (self.sun_distance + focus_size))
                self.cam_node.look_at(focus_point)

                snap_shadow_map(self.mvp, self.cam_node, self.resolution)
                self.target.active = True
                self.pta_mvp[0] = self.mvp

                self.focus = None
        else:
            self.target.active = False

    def create(self):
        self.camera = Camera("PSSMSceneSunShadowCam")
        self.cam_lens = OrthographicLens()
        self.cam_lens.set_film_size(400, 400)
        self.cam_lens.set_near_far(100.0, 1800.0)
        self.camera.set_lens(self.cam_lens)
        self.cam_node = Globals.base.render.attach_new_node(self.camera)

        self.target = self.create_target("ShadowMap")
        self.target.size = self.resolution
        self.target.add_depth_attachment(bits=32)
        self.target.prepare_render(self.cam_node)

        # Register shadow camera
        self._pipeline.tag_mgr.register_camera("shadow", self.camera)

    def set_shader_input(self, *args):
        Globals.render.set_shader_input(*args)

    def set_shader_inputs(self, **kwargs):
        Globals.render.set_shader_inputs(**kwargs)



import math

import numpy as np

from metadrive.component.block.base_block import BaseBlock
from metadrive.component.lane.scenario_lane import ScenarioLane
from metadrive.component.road_network.edge_road_network import EdgeRoadNetwork
from metadrive.constants import DrivableAreaProperty
from metadrive.constants import PGLineType, PGLineColor
from metadrive.engine.engine_utils import get_engine
from metadrive.scenario.scenario_description import ScenarioDescription
from metadrive.type import MetaDriveType
from metadrive.utils.coordinates_shift import panda_heading
from metadrive.utils.interpolating_line import InterpolatingLine
from metadrive.utils.math import norm


class ScenarioBlock(BaseBlock):
    LINE_CULL_DIST = 500

    def __init__(self, block_index: int, global_network, random_seed, map_index, need_lane_localization):
        # self.map_data = map_data
        self.need_lane_localization = need_lane_localization
        self.map_index = map_index
        data = self.engine.data_manager.current_scenario
        sdc_track = data.get_sdc_track()
        self.sdc_start_point = sdc_track["state"]["position"][0]
        super(ScenarioBlock, self).__init__(block_index, global_network, random_seed)

    @property
    def map_data(self):
        e = get_engine()
        return e.data_manager.get_scenario(self.map_index, should_copy=False)["map_features"]

    def _sample_topology(self) -> bool:
        for lane_id, data in self.map_data.items():
            if MetaDriveType.is_lane(data.get("type", False)):
                if len(data[ScenarioDescription.POLYLINE]) <= 1:
                    continue
                lane = ScenarioLane(lane_id, self.map_data, self.need_lane_localization)
                self.block_network.add_lane(lane)
        return True

    def create_in_world(self):
        """
        The lane line should be created separately
        """
        graph = self.block_network.graph
        for id, lane_info in graph.items():
            lane = lane_info.lane
            lane.construct_lane_in_block(self, lane_index=id)
            # lane.construct_lane_line_in_block(self, [True if len(lane.left_lanes) == 0 else False,
            #                                          True if len(lane.right_lanes) == 0 else False, ])
        # draw
        for lane_id, data in self.map_data.items():
            type = data.get("type", None)
            if ScenarioDescription.POLYLINE in data and len(data[ScenarioDescription.POLYLINE]) <= 1:
                continue
            if MetaDriveType.is_road_line(type):
                if MetaDriveType.is_broken_line(type):
                    self.construct_broken_line(
                        np.asarray(data[ScenarioDescription.POLYLINE]),
                        PGLineColor.YELLOW if MetaDriveType.is_yellow_line(type) else PGLineColor.GREY
                    )
                else:
                    self.construct_continuous_line(
                        np.asarray(data[ScenarioDescription.POLYLINE]),
                        PGLineColor.YELLOW if MetaDriveType.is_yellow_line(type) else PGLineColor.GREY
                    )
            # elif MetaDriveType.is_road_edge(type) and MetaDriveType.is_sidewalk(type):
            #     self.construct_sidewalk(
            #         convert_polyline_to_metadrive(
            #             data[ScenarioDescription.POLYLINE], coordinate_transform=self.coordinate_transform
            #         )
            #     )
            # elif MetaDriveType.is_road_edge(type) and not MetaDriveType.is_sidewalk(type):
            #     self.construct_continuous_line(
            #         convert_polyline_to_metadrive(
            #             data[ScenarioDescription.POLYLINE], coordinate_transform=self.coordinate_transform
            #         ), PGLineColor.GREY
            #     )
            # else:
            #     raise ValueError("Can not build lane line type: {}".format(type))
            # TODO LQY: DO we need sidewalk?
            elif MetaDriveType.is_road_edge(type):
                self.construct_continuous_line(np.asarray(data[ScenarioDescription.POLYLINE]), color=PGLineColor.GREY)

    def construct_continuous_line(self, polyline, color):
        line = InterpolatingLine(polyline)
        segment_num = int(line.length / DrivableAreaProperty.STRIPE_LENGTH)
        for segment in range(segment_num):
            start = line.get_point(DrivableAreaProperty.STRIPE_LENGTH * segment)
            # trick for optimizing
            dist = norm(start[0] - self.sdc_start_point[0], start[1] - self.sdc_start_point[1])
            if dist > self.LINE_CULL_DIST:
                continue

            if segment == segment_num - 1:
                end = line.get_point(line.length)
            else:
                end = line.get_point((segment + 1) * DrivableAreaProperty.STRIPE_LENGTH)
            node_path_list = ScenarioLane.construct_lane_line_segment(self, start, end, color, PGLineType.CONTINUOUS)
            self._node_path_list.extend(node_path_list)

    def construct_broken_line(self, polyline, color):
        line = InterpolatingLine(polyline)
        segment_num = int(line.length / (2 * DrivableAreaProperty.STRIPE_LENGTH))
        for segment in range(segment_num):
            start = line.get_point(segment * DrivableAreaProperty.STRIPE_LENGTH * 2)
            # trick for optimizing
            dist = norm(start[0] - self.sdc_start_point[0], start[1] - self.sdc_start_point[1])
            if dist > self.LINE_CULL_DIST:
                continue
            end = line.get_point(segment * DrivableAreaProperty.STRIPE_LENGTH * 2 + DrivableAreaProperty.STRIPE_LENGTH)
            if segment == segment_num - 1:
                end = line.get_point(line.length - DrivableAreaProperty.STRIPE_LENGTH)
            node_path_list = ScenarioLane.construct_lane_line_segment(self, start, end, color, PGLineType.BROKEN)
            self._node_path_list.extend(node_path_list)

    def construct_sidewalk(self, polyline):
        line = InterpolatingLine(polyline)
        seg_len = DrivableAreaProperty.LANE_SEGMENT_LENGTH
        segment_num = int(line.length / seg_len)
        # last_theta = None
        for segment in range(segment_num):
            lane_start = line.get_point(segment * seg_len)
            lane_end = line.get_point((segment + 1) * seg_len)
            if segment == segment_num - 1:
                lane_end = line.get_point(line.length)
            direction_v = lane_end - lane_start
            theta = panda_heading(math.atan2(direction_v[1], direction_v[0]))
            # if segment == segment_num - 1:
            #     factor = 1
            # else:
            #     factor = 1.25
            #     if last_theta is not None:
            #         diff = wrap_to_pi(theta) - wrap_to_pi(last_theta)
            #         if diff > 0:
            #             factor += math.sin(abs(diff) / 2) * DrivableAreaProperty.SIDEWALK_WIDTH / norm(
            #                 lane_start[0] - lane_end[0], lane_start[1] - lane_end[1]
            #             ) + 0.15
            #         else:
            #             factor -= math.sin(abs(diff) / 2) * DrivableAreaProperty.SIDEWALK_WIDTH / norm(
            #                 lane_start[0] - lane_end[0], lane_start[1] - lane_end[1]
            #             )
            last_theta = theta
            node_path_list = ScenarioLane.construct_sidewalk_segment(
                self, lane_start, lane_end, length_multiply=1, extra_thrust=0, width=0.2
            )
            self._node_path_list.extend(node_path_list)

    @property
    def block_network_type(self):
        return EdgeRoadNetwork

    def destroy(self):
        self.map_index = None
        # self.map_data = None
        super(ScenarioBlock, self).destroy()

    def __del__(self):
        self.destroy()
        super(ScenarioBlock, self).__del__()



import copy

import numpy as np
from scipy.optimize import minimize


class PhysicSetting:
    """
    Work in progress
    """
    def __init__(self, size, hpr, offset, model_path):
        # offset is used for 4 wheels.
        self.size = size
        self.hpr = hpr
        self.offset = offset
        self.model_path = model_path

    @classmethod
    def offset_after_skew(cls, ps, dx, dy, index, model_path):
        # this is designed for vehicles' wheels,
        # so you dont have to repeat PhysicSetting.__init__(...) 4 times :)

        # 0 - right front wheel
        # 1 - left front wheel
        # 2 - right rear wheel
        # 3 - left rear wheel
        os = ps.offset
        new_os = os + dx * ((1 & index) * 2 - 1) + dy * (1 if index < 2 else -1)
        new_ps = PhysicSetting(ps.size, ps.hpr, new_os, model_path)
        return new_ps


class ModelPredictiveControl:
    """
    Work in progress
    """
    def __init__(self, horizon, dim, dt):
        self.state = None
        self.target = None
        self.horizon = horizon
        self.dim = dim
        self.dt = dt
        self.mass = 800
        self.len = 1
        self.bounds = []
        self.u = np.zeros(self.dim * self.horizon, dtype=np.float32)
        self.config = {"replan": False}

    def cost(self, u, *args):
        raise NotImplementedError

    def plant_model(self, state, dt, *control):
        raise NotImplementedError

    def solve(self):
        if self.config['replan']:
            self.u = np.zeros(self.dim * self.horizon, dtype=np.float32)
        else:
            for _ in range(self.dim):
                self.u = np.delete(self.u, 0)
            for _ in range(self.dim):
                self.u = np.append(self.u, self.u[-1 * self.dim])


class OpponentModelPredictiveControl(ModelPredictiveControl):
    def __init__(self, landmarks, horizon, dim, repeat, dt):
        super(OpponentModelPredictiveControl, self).__init__(horizon, dim, dt)
        self.landmarks = landmarks
        self.target_index = 0
        self.dt = dt
        self.dim = dim
        self.repeat = repeat

        for i in range(self.horizon):
            for _ in range(1):
                self.bounds += [[-1, 1]]
            for _ in range(1):
                self.bounds += [[-1, 1]]

    def cost(self, u, *args):
        state = args[0]
        vehicles = args[1]
        cost = 0.0
        for i in range(self.horizon - 1):
            head_u = u[self.dim * i:(self.dim) * (i + 1)]
            tail_u = u[self.dim * (i + 1):(self.dim) * (i + 2)]
            for j in range(self.repeat):
                state = self.plant_model(
                    state, self.dt, [head_u[_] + (tail_u[_] - head_u[_] * (j / self.repeat)) for _ in range(self.dim)]
                )
                x, y = state[0], state[1]
                dist = (
                    (x - self.landmarks[self.target_index][0])**2 + (y - self.landmarks[self.target_index][1])**2
                )**.5
                cost += dist
                for vehicle in vehicles:

                    p = np.array(
                        [
                            x - vehicle.position[0] - self.dt * (i + j / self.repeat) * vehicle.velocity[0],
                            y + vehicle.position[1]
                        ]
                    )
                    d = (p[0] - x)**2 + (p[1] + y)**2
                    alpha = -vehicle.heading
                    c, s = math.cos(alpha), math.sin(alpha)
                    ratio = 4
                    # print(alpha)
                    d = p @ np.array(
                        [[c**2 + ratio * s**2, -(ratio - 1) * c * s], [-(ratio - 1) * c * s, c**2 + ratio * s**2]]
                    ) @ p.T
                    if d > 50:
                        cost += 20
                    else:
                        cost += 1000 / d

            if state[2] > 6:
                cost += (state[2] - 6) * 10
            # if state[2] < 0.1:
            #     cost += 10000
            if state[2] > 1:
                cost += np.std(u[1::2]) * 20

        # print('current cost %.4f' % cost)
        # cost += (state[2] - 10) **2
        return cost

    def plant_model(self, state, dt, *control):
        [x, y, v, phi, beta] = state
        pedal, steering = control[0]
        pedal *= 500
        steering *= 40
        new_beta = np.arctan(0.5 * np.tan(steering / 180 * np.pi))

        if pedal < 0:
            pedal *= 3

        a = pedal / 500 * 3

        af = .5

        new_v = 0
        if v > 1e-5 or a > af:
            new_v = v + (a - af) * dt
            if v * new_v < 0:
                v = new_v = 0

        new_phi = phi + v / 2 / 2 * np.tan(steering / 180 * np.pi) * dt
        new_x = x + v * math.cos(phi + beta) * dt
        new_y = y + v * math.sin(phi + beta) * dt
        new_state = [new_x, new_y, new_v, new_phi, new_beta]
        return new_state

    def update_target(self, pos):
        dist = lambda pos, lm, i: ((pos[0] - lm[i][0])**2 + (pos[1] - lm[i][1])**2)**0.5

        if self.target_index == 0:
            while dist(pos, self.landmarks, self.target_index) > dist(pos, self.landmarks, self.target_index + 1):
                self.target_index += 1

        while (self.landmarks[self.target_index][0] - pos[0]) * \
                (self.landmarks[self.target_index + 1][0] - pos[0]) + \
                (self.landmarks[self.target_index][1] - pos[1]) * \
                (self.landmarks[self.target_index + 1][1] - pos[1]) < 0:
            self.target_index += 1
        d = dist(pos, self.landmarks, self.target_index)
        if d < 5:
            self.target_index += 1
        return self.target_index

    def solve(self, ref, vehicles):
        new_u = copy.deepcopy(self.u)
        for _ in range(self.dim):
            new_u = np.delete(new_u, 0)
        for _ in range(self.dim):
            new_u = np.append(new_u, new_u[-1 * self.dim])
        for _ in range(self.dim * self.horizon):
            self.u[_] += (new_u[_] - self.u[_]) / self.repeat
        for _ in range(self.dim * self.horizon):
            self.u[_] = min(max(self.u[_], self.bounds[_][0]), self.bounds[_][1])

        # self.u = np.zeros(self.dim * self.horizon)

        u_optimze = minimize(
            self.cost, self.u, (ref, vehicles), method="SLSQP", bounds=self.bounds, tol=1e-5, options={'disp': False}
        )
        self.u = u_optimze.x
        return self.u[:self.dim]



from metadrive.component.map.base_map import BaseMap
from metadrive.component.map.pg_map import MapGenerateMethod
from metadrive.envs.safe_metadrive_env import SafeMetaDriveEnv
from metadrive.manager.traffic_manager import TrafficMode
from metadrive.policy.idm_policy import IDMPolicy
from metadrive.utils import setup_logger
from metadrive.utils.math import norm


def assert_equal_pos(data_1, data_2):
    assert len(data_1) == len(data_2)
    for i in range(len(data_1)):
        difference = data_1[i] - data_2[i]
        diff = norm(data_1[i][0] - data_2[i][0], data_1[i][1] - data_2[i][1])
        assert diff < 0.00001, "pos mismatch for vehicle: {}, distance: {}".format(i, diff)


def test_scenario_randomness(vis=False):
    setup_logger(True)
    cfg = {
        "accident_prob": 0.8,
        "num_scenarios": 1,
        "traffic_density": 0.1,
        "start_seed": 1000,
        "manual_control": False,
        "use_render": vis,
        "agent_policy": IDMPolicy,
        "traffic_mode": TrafficMode.Trigger,
        "record_episode": False,
        "map_config": {
            BaseMap.GENERATE_TYPE: MapGenerateMethod.BIG_BLOCK_SEQUENCE,
            BaseMap.GENERATE_CONFIG: "CrCSC",
            BaseMap.LANE_WIDTH: 3.5,
            BaseMap.LANE_NUM: 3,
        }
    }

    env = SafeMetaDriveEnv(cfg)
    try:
        positions_1 = []
        o, _ = env.reset()
        positions_1.append([env.vehicle.position] + [v.position for v in env.engine.traffic_manager.traffic_vehicles])
        for i in range(1, 100000 if vis else 2000):
            o, r, tm, tc, info = env.step([0, 1])
            positions_1.append(
                [env.vehicle.position] + [v.position for v in env.engine.traffic_manager.traffic_vehicles]
            )
            if tm or tc:
                break
        env.close()
        positions_1.reverse()
        env = SafeMetaDriveEnv(cfg)
        o, _ = env.reset()
        old_position = positions_1.pop()
        new_position = [env.vehicle.position] + [v.position for v in env.engine.traffic_manager.traffic_vehicles]
        assert_equal_pos(old_position, new_position)
        for i in range(1, 100000 if vis else 2000):
            o, r, tm, tc, info = env.step([0, 1])
            old_position = positions_1.pop()
            new_position = [env.vehicle.position] + [v.position for v in env.engine.traffic_manager.traffic_vehicles]
            assert_equal_pos(old_position, new_position)
            if tm or tc:
                break
    finally:
        env.close()


if __name__ == "__main__":
    test_scenario_randomness(vis=True)



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from panda3d.egg import (EggGroup, EggPolygon, EggVertexPool, EggMaterial, EggTexture)

from metadrive.third_party.kitsunetsuki.base.matrices import get_object_matrix
from metadrive.third_party.kitsunetsuki.base.armature import get_armature
from metadrive.third_party.kitsunetsuki.base.mesh import obj2mesh
from metadrive.third_party.kitsunetsuki.base.objects import apply_modifiers, is_collision


class GeomMixin(object):
    def _get_joints(self, egg_group):
        results = {}

        for child in egg_group.get_children():
            if (isinstance(child, EggGroup) and child.get_group_type() == EggGroup.GT_joint):
                results[child.get_name()] = child
                results.update(self._get_joints(child))

        return results

    def make_geom(self, node, obj, can_merge=False):
        triangulate = not is_collision(obj)
        if self._geom_scale != 1:
            obj.scale.x = self._geom_scale
            obj.scale.y = self._geom_scale
            obj.scale.z = self._geom_scale
            apply_modifiers(obj, triangulate=triangulate, apply_scale=True)
        else:
            apply_modifiers(obj, triangulate=triangulate)
        mesh = obj2mesh(obj, triangulate=triangulate)

        # get or create materials and textures
        egg_materials = {}
        egg_textures = {}
        egg_material_textures = {}
        if not self._no_materials and not is_collision(obj):
            for material in mesh.materials.values():
                # material
                for child in self._root.get_children():  # existing material
                    if (isinstance(child, EggMaterial) and child.get_name() == material.name):
                        egg_materials[material.name] = child
                        break
                else:  # new material
                    egg_material = self.make_material(material)
                    self._root.add_child(egg_material)
                    egg_materials[material.name] = egg_material

                # material -> textures
                if material.name not in egg_material_textures:
                    egg_material_textures[material.name] = {}

                # textures
                if not self._no_textures:
                    for type_, _, egg_texture in self.make_textures(material):
                        tname = egg_texture.get_name()
                        for child in self._root.get_children():  # existing texture
                            if (isinstance(child, EggTexture) and child.get_name() == tname):
                                egg_textures[tname] = child
                                egg_material_textures[material.name][tname] = child
                                break
                        else:  # new texture
                            self._root.add_child(egg_texture)
                            egg_textures[tname] = egg_texture
                            egg_material_textures[material.name][tname] = egg_texture

        # get or create vertex pool
        egg_vertex_pool = None
        egg_vertex_id = 0
        if can_merge:
            for child in node.get_children():  # existing vertex pool
                if isinstance(child, EggVertexPool):
                    egg_vertex_pool = child
                    egg_vertex_id = egg_vertex_pool.get_highest_index()
                    break

        if egg_vertex_pool is None:  # new vertex pool
            egg_vertex_pool = EggVertexPool(node.get_name())
            egg_vertex_id = egg_vertex_pool.get_highest_index()
            node.add_child(egg_vertex_pool)

        # get armature and joints
        armature = get_armature(obj)
        egg_joints = {}
        if armature:
            for child in self._root.get_children():
                if (isinstance(child, EggGroup) and child.get_dart_type() == EggGroup.DT_structured
                        and child.get_name() == armature.name):
                    egg_joints = self._get_joints(child)

        sharp_vertices = {}
        uv_tb = {}
        if not is_collision(obj):
            sharp_vertices = self.get_sharp_vertices(mesh)
            uv_tb = self.get_tangent_bitangent(mesh)
        egg_vertices = {}

        obj_matrix = get_object_matrix(obj, armature)
        parent_obj_matrix = obj_matrix
        if armature:
            parent_obj_matrix = get_object_matrix(armature)

        for polygon in mesh.polygons:
            # <-- polygon
            material = None
            mname = None
            if not self._no_materials:
                try:
                    material = mesh.materials[polygon.material_index]
                    mname = material.name
                except IndexError:
                    pass

            # make polygon
            egg_polygon = EggPolygon(mname or node.get_name())

            # set material and textures
            if material and not self._no_materials and not is_collision(obj):
                if mname in egg_materials:
                    egg_polygon.set_material(egg_materials[mname])

                # set textures
                if mname in egg_material_textures and not self._no_textures:
                    for egg_texture in egg_material_textures[mname].values():
                        egg_polygon.add_texture(egg_texture)

            # vertices
            for i, vertex_id in enumerate(polygon.vertices):
                # i is vertex counter inside a polygon
                # (0, 1, 2) for triangle
                # vertex_id is reusable id,
                # because multiple polygons can share the same vertices

                # <-- vertex
                vertex = mesh.vertices[vertex_id]
                use_smooth = (polygon.use_smooth and vertex_id not in sharp_vertices and not is_collision(obj))

                # try to reuse shared vertices
                if (polygon.use_smooth and vertex_id in egg_vertices and not is_collision(obj)):
                    shared = False
                    for egg_vertex in egg_vertices[vertex_id]:
                        loop_id = polygon.loop_indices[i]
                        egg_vertex_uv = egg_vertex.get_uv_obj(self._get_uv_name(mesh.uv_layers.active))

                        if not egg_vertex_uv:
                            egg_polygon.add_vertex(egg_vertex)
                            shared = True
                            break

                        if self.can_share_vertex(mesh, vertex, loop_id, egg_vertex_uv.get_uv(),
                                                 egg_vertex.get_normal()):
                            egg_polygon.add_vertex(egg_vertex)
                            shared = True
                            break
                    if shared:
                        continue

                # make new vertex data
                can_merge_vertices = can_merge
                if armature:
                    can_merge_vertices = True
                elif is_collision(obj):
                    can_merge_vertices = False
                egg_vertex = self.make_vertex(
                    parent_obj_matrix, obj_matrix, polygon, vertex, use_smooth=use_smooth, can_merge=can_merge_vertices
                )

                # uv layers
                if not is_collision(obj):
                    for uv_name, uv_layer in mesh.uv_layers.items():
                        # <-- vertex uv
                        loop_id = polygon.loop_indices[i]
                        uv_loop = uv_layer.data[loop_id]

                        # not active layer and extra UV disabled
                        if not uv_layer.active and self._no_extra_uv:
                            continue

                        egg_vertex_uv = self.make_vertex_uv(uv_layer, uv_loop.uv)
                        if uv_name in uv_tb:
                            t, b, s = uv_tb[uv_name][loop_id]
                            tangent = parent_obj_matrix @ t
                            binormal = parent_obj_matrix @ b
                            egg_vertex_uv.set_tangent(tuple(tangent))
                            egg_vertex_uv.set_binormal(tuple(binormal))
                        egg_vertex.set_uv_obj(egg_vertex_uv)
                        # vertex uv -->

                # generate new ID, add vertex and save last ID
                egg_vertex_id += 1
                egg_vertex_pool.add_vertex(egg_vertex, egg_vertex_id)
                egg_vertex_pool.set_highest_index(egg_vertex_id)
                egg_polygon.add_vertex(egg_vertex)

                # save vertex data for sharing
                if vertex_id not in egg_vertices:
                    egg_vertices[vertex_id] = []
                egg_vertices[vertex_id].append(egg_vertex)

                # attach joints to vertex
                if armature:
                    for vertex_group in vertex.groups:
                        obj_vertex_group = obj.vertex_groups[vertex_group.group]
                        if obj_vertex_group.name in egg_joints:
                            egg_joint = egg_joints[obj_vertex_group.name]
                            egg_joint.set_vertex_membership(egg_vertex, vertex_group.weight)

                # vertex -->

            node.add_child(egg_polygon)
            # polygon -->



from metadrive.component.buildings.tollgate_building import TollGateBuilding
from metadrive.component.pgblock.bottleneck import PGBlock
from metadrive.component.pgblock.create_pg_block_utils import CreateAdverseRoad, CreateRoadFrom, ExtendStraightLane
from metadrive.component.pgblock.pg_block import PGBlockSocket
from metadrive.component.road_network import Road
from metadrive.constants import PGLineType, PGLineColor
from metadrive.engine.engine_utils import get_engine
from metadrive.component.pg_space import ParameterSpace, Parameter, BlockParameterSpace


class TollGate(PGBlock):
    """
    Toll, like Straight, but has speed limit
    """
    SOCKET_NUM = 1
    PARAMETER_SPACE = ParameterSpace(BlockParameterSpace.BOTTLENECK_PARAMETER)
    ID = "$"

    SPEED_LIMIT = 3  # m/s ~= 5 miles per hour https://bestpass.com/feed/61-speeding-through-tolls

    def _try_plug_into_previous_block(self) -> bool:
        self.set_part_idx(0)  # only one part in simple block like straight, and curve
        para = self.get_config()
        length = para[Parameter.length]
        self.BUILDING_LENGTH = length
        basic_lane = self.positive_basic_lane
        new_lane = ExtendStraightLane(basic_lane, length, [PGLineType.CONTINUOUS, PGLineType.SIDE])
        start = self.pre_block_socket.positive_road.end_node
        end = self.add_road_node()
        socket = Road(start, end)
        _socket = -socket

        # create positive road
        no_cross = CreateRoadFrom(
            new_lane,
            self.positive_lane_num,
            socket,
            self.block_network,
            self._global_network,
            center_line_color=PGLineColor.YELLOW,
            center_line_type=PGLineType.CONTINUOUS,
            inner_lane_line_type=PGLineType.CONTINUOUS,
            side_lane_line_type=PGLineType.SIDE,
            ignore_intersection_checking=self.ignore_intersection_checking
        )

        # create negative road
        no_cross = CreateAdverseRoad(
            socket,
            self.block_network,
            self._global_network,
            center_line_color=PGLineColor.YELLOW,
            center_line_type=PGLineType.CONTINUOUS,
            inner_lane_line_type=PGLineType.CONTINUOUS,
            side_lane_line_type=PGLineType.SIDE,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross

        self.add_sockets(PGBlockSocket(socket, _socket))
        self._add_building_and_speed_limit(socket)
        self._add_building_and_speed_limit(_socket)
        return no_cross

    def _add_building_and_speed_limit(self, road):
        # add house
        lanes = road.get_lanes(self.block_network)
        for idx, lane in enumerate(lanes):
            lane.set_speed_limit(self.SPEED_LIMIT)
            if idx % 2 == 1:
                # add toll
                position = lane.position(lane.length / 2, 0)
                building = get_engine().spawn_object(
                    TollGateBuilding, lane=lane, position=position, heading_theta=lane.heading_theta_at(0)
                )
                self.dynamic_nodes.append(building.body)
                self._block_objects.append(building)



from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.pgblock.ramp import OutRampOnStraight
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.engine.asset_loader import initialize_asset_loader
from metadrive.tests.vis_block.vis_block_base import TestBlock

if __name__ == "__main__":
    test = TestBlock()

    initialize_asset_loader(test)

    global_network = NodeRoadNetwork()
    straight = FirstPGBlock(global_network, 3.0, 1, test.render, test.world, 1)
    for i in range(1, 3):
        straight = OutRampOnStraight(i, straight.get_socket(0), global_network, i)
        # print(straight.construct_block(test.render, test.world))
        # print(len(straight.dynamic_nodes))
    test.show_bounding_box(global_network)
    test.run()



import numpy as np
from metadrive.component.traffic_participants.pedestrian import Pedestrian
from metadrive.utils.math import wrap_to_pi

from metadrive.envs.metadrive_env import MetaDriveEnv


def test_set_get_vehicle_attribute(render=False):
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.,
            "decision_repeat": 1,
            "map": "SSS",
            "use_render": render,
        }
    )
    try:
        o, _ = env.reset()
        for _ in range(10):
            env.vehicle.set_velocity([5, 0], in_local_frame=False)
            o, r, tm, tc, info = env.step([0, 0])
            assert abs(env.vehicle.speed - 5) < 0.01  # may encounter friction
            assert np.isclose(env.vehicle.velocity, np.array([5, 0]), rtol=1e-2, atol=1e-2).all()
            assert abs(env.vehicle.speed - env.vehicle.speed_km_h / 3.6) < 1e-4
            assert np.isclose(env.vehicle.velocity, env.vehicle.velocity_km_h / 3.6).all()

        for _ in range(10):
            o, r, tm, tc, info = env.step([0, 0])
            env.vehicle.set_velocity([0, 5], in_local_frame=False)
            assert abs(env.vehicle.speed - 5) < 0.1
            assert np.isclose(env.vehicle.velocity, np.array([0, 5]), rtol=1e-5, atol=1e-5).all()
            assert abs(env.vehicle.speed - env.vehicle.speed_km_h / 3.6) < 1e-4
            assert np.isclose(env.vehicle.velocity, env.vehicle.velocity_km_h / 3.6).all()

        for _ in range(10):
            o, r, tm, tc, info = env.step([0, 0])
            env.vehicle.set_velocity([5, 3], value=10, in_local_frame=False)
            assert abs(env.vehicle.speed - 10) < 0.1
            assert np.isclose(
                env.vehicle.velocity,
                np.array([5 / np.linalg.norm(np.array([5, 3])) * 10, 3 / np.linalg.norm(np.array([5, 3])) * 10]),
                rtol=1e-5,
                atol=1e-5
            ).all()
            assert abs(env.vehicle.speed - env.vehicle.speed_km_h / 3.6) < 1e-4
            assert np.isclose(env.vehicle.velocity, env.vehicle.velocity_km_h / 3.6).all()

        for _ in range(10):
            o, r, tm, tc, info = env.step([0, 0])
            env.vehicle.set_velocity([0.3, 0.1], value=10, in_local_frame=False)
            assert abs(env.vehicle.speed - 10) < 0.1
            assert np.isclose(
                env.vehicle.velocity,
                np.array(
                    [0.3 / np.linalg.norm(np.array([0.3, 0.1])) * 10, 0.1 / np.linalg.norm(np.array([0.3, 0.1])) * 10]
                ),
                rtol=1e-5,
                atol=1e-5
            ).all()
            assert abs(env.vehicle.speed - env.vehicle.speed_km_h / 3.6) < 0.0001
            assert np.isclose(env.vehicle.velocity, env.vehicle.velocity_km_h / 3.6).all()

    finally:
        env.close()


def test_coordinates(render=False):
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.,
            "decision_repeat": 1,
            "vehicle_config": {
                "enable_reverse": True
            },
            "map": "SSS",
            "use_render": render,
        }
    )
    try:
        o, _ = env.reset()
        assert abs(env.vehicle.heading_theta) == 0
        assert np.isclose(env.vehicle.heading, [1.0, 0]).all()
        env.vehicle.set_velocity([5, 0], in_local_frame=True)
        for _ in range(10):
            env.vehicle.set_velocity([5, 0], in_local_frame=True)
            o, r, tm, tc, info = env.step([0, 0])
        assert abs(env.vehicle.velocity[0] - 5.) < 1e-2 and abs(env.vehicle.velocity[1]) < 0.001

        o, _ = env.reset()
        assert abs(env.vehicle.heading_theta) == 0
        assert np.isclose(env.vehicle.heading, [1.0, 0]).all()
        env.vehicle.set_velocity([5, 0], in_local_frame=False)
        for _ in range(10):
            o, r, tm, tc, info = env.step([0, 0])
        assert env.vehicle.velocity[0] > 3. and abs(env.vehicle.velocity[1]) < 0.001

        env.reset()
        env.vehicle.set_velocity([0, 5], in_local_frame=False)

        for _ in range(1):
            o, r, tm, tc, info = env.step([0, 0])
        assert env.vehicle.velocity[1] > 3. and abs(env.vehicle.velocity[0]) < 0.002

        env.reset()
        assert abs(env.vehicle.heading_theta) == 0
        assert np.isclose(env.vehicle.heading, [1.0, 0]).all()
        env.vehicle.set_velocity([-5, 0], in_local_frame=False)
        for _ in range(10):
            o, r, tm, tc, info = env.step([0, 0])
        assert env.vehicle.velocity[0] < -3. and abs(env.vehicle.velocity[1]) < 0.001

        env.vehicle.set_velocity([0, -5], in_local_frame=False)

        for _ in range(1):
            o, r, tm, tc, info = env.step([0, 0])
        assert env.vehicle.velocity[1] < -3. and abs(env.vehicle.velocity[0]) < 0.002

        # steering left
        env.reset()
        begining_pos = env.vehicle.position
        assert abs(env.vehicle.heading_theta) == 0
        assert np.isclose(env.vehicle.heading, [1.0, 0]).all()

        for _ in range(100):
            o, r, tm, tc, info = env.step([0.8, 0.8])
        assert env.vehicle.velocity[1] > 1. and abs(env.vehicle.velocity[0]) > 1
        assert env.vehicle.heading_theta > 0.3  # rad
        assert env.vehicle.position[0] > begining_pos[0] and env.vehicle.position[1] > begining_pos[1]

        # steering right
        env.reset()
        begining_pos = env.vehicle.position
        assert abs(env.vehicle.heading_theta) == 0
        assert np.isclose(env.vehicle.heading, [1.0, 0]).all()

        for _ in range(100):
            o, r, tm, tc, info = env.step([-0.8, 0.8])
        assert env.vehicle.velocity[1] < -1. and abs(env.vehicle.velocity[0]) > 1
        assert env.vehicle.position[0] > begining_pos[0] and env.vehicle.position[1] < begining_pos[1]
        assert env.vehicle.heading_theta < -0.3  # rad

        env.reset()
        env.vehicle.set_heading_theta(np.deg2rad(90))
        for _ in range(10):
            o, r, tm, tc, info, = env.step([-0., 0.])
        assert wrap_to_pi(abs(env.vehicle.heading_theta - np.deg2rad(90))) < 1
        assert np.isclose(env.vehicle.heading, np.array([0, 1]), 1e-4, 1e-4).all()

        env.reset()
        env.vehicle.set_heading_theta(np.deg2rad(45))
        for _ in range(10):
            o, r, tm, tc, info, = env.step([-0., 0.])
        assert wrap_to_pi(abs(env.vehicle.heading_theta - np.deg2rad(45))) < 1
        assert np.isclose(env.vehicle.heading, np.array([np.sqrt(2) / 2, np.sqrt(2) / 2]), 1e-4, 1e-4).all()

        env.reset()
        env.vehicle.set_heading_theta(np.deg2rad(-90))
        for _ in range(10):
            o, r, tm, tc, info, = env.step([-0., 0.])
        assert abs(env.vehicle.heading_theta + np.deg2rad(-90) + np.pi) < 0.01
        assert np.isclose(env.vehicle.heading, np.array([0, -1]), 1e-4, 1e-4).all()
    finally:
        env.close()


def test_set_angular_v_and_set_v_no_friction(render=False):
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.,
            "decision_repeat": 5,
            "map": "SSS",
            "use_render": render,
            "vehicle_config": {
                "no_wheel_friction": True
            }
        }
    )
    try:
        o, _ = env.reset()
        for _ in range(100):
            # 10 s , np.pi/10 per second
            env.vehicle.set_angular_velocity(np.pi / 10)
            o, r, tm, tc, info = env.step([0, 0])
        assert abs(wrap_to_pi(env.vehicle.heading_theta) - np.pi) < 1e-2, env.vehicle.heading_theta
        # print(env.vehicle.heading_theta / np.pi * 180)

        o, _ = env.reset()
        for _ in range(100):
            # 10 s , np.pi/10 per second
            env.vehicle.set_angular_velocity(18, in_rad=False)
            o, r, tm, tc, info = env.step([0, 0])
        assert abs(wrap_to_pi(env.vehicle.heading_theta) - np.pi) < 1e-2, env.vehicle.heading_theta
        # print(env.vehicle.heading_theta / np.pi * 180)

        o, _ = env.reset()
        start = env.vehicle.position[0]
        for _ in range(100):
            # 10 s
            env.vehicle.set_velocity([1, 0], in_local_frame=True)
            o, r, tm, tc, info = env.step([0, 0])
        assert abs(env.vehicle.position[0] - start - 10) < 5e-2, env.vehicle.position

        o, _ = env.reset()
        start = env.vehicle.position[0]
        for _ in range(100):
            # 10 s
            env.vehicle.set_velocity([1, 0], in_local_frame=False)
            o, r, tm, tc, info = env.step([0, 0])
        assert abs(env.vehicle.position[0] - start - 10) < 5e-2, env.vehicle.position

        o, _ = env.reset()
        start = env.vehicle.position[1]
        for _ in range(10):
            # 10 s
            env.vehicle.set_velocity([0, 1], in_local_frame=False)
            o, r, tm, tc, info = env.step([0, 0])
        assert abs(env.vehicle.position[1] - start - 1) < 5e-2, env.vehicle.position

        o, _ = env.reset()
        start = env.vehicle.position[0]
        env.vehicle.set_heading_theta(-np.pi / 2)
        for _ in range(100):
            # 10 s
            env.vehicle.set_velocity([0, 1], in_local_frame=True)
            o, r, tm, tc, info = env.step([0, 0])
        assert abs(env.vehicle.position[0] - start - 10) < 5e-2, env.vehicle.position

        o, _ = env.reset()
        start = env.vehicle.position[0]
        env.vehicle.set_heading_theta(-np.pi / 2)
        for _ in range(100):
            # 10 s
            env.vehicle.set_velocity([1, 0], in_local_frame=False)
            o, r, tm, tc, info = env.step([0, 0])
        assert abs(env.vehicle.position[0] - start - 10) < 5e-2, env.vehicle.position
    finally:
        env.close()


def test_set_angular_v_and_set_v_no_friction_pedestrian(render=False):
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.,
            "decision_repeat": 5,
            "map": "S",
            "use_render": render,
            "vehicle_config": {
                "no_wheel_friction": True
            }
        }
    )
    try:
        o, _ = env.reset()
        env.engine.terrain.dynamic_nodes[0].setFriction(0.)
        obj_1 = env.engine.spawn_object(Pedestrian, position=[10, 3], heading_theta=0, random_seed=1)
        for _ in range(10):
            # 10 s , np.pi/10 per second
            obj_1.set_angular_velocity(np.pi / 10)
            o, r, tm, tc, info = env.step([0, 0])
        assert abs(wrap_to_pi(obj_1.heading_theta) - np.pi / 10) < 1e-2, obj_1.heading_theta
        obj_1.destroy()

        o, _ = env.reset()
        env.engine.terrain.dynamic_nodes[0].setFriction(0.)
        obj_1 = env.engine.spawn_object(Pedestrian, position=[10, 3], heading_theta=0, random_seed=1)
        for _ in range(10):
            # obj_1.set_position([30,0], 10)
            # 10 s , np.pi/10 per second
            obj_1.set_angular_velocity(18, in_rad=False)
            o, r, tm, tc, info = env.step([0, 0])
        assert abs(wrap_to_pi(obj_1.heading_theta) - np.pi / 10) < 1e-2, obj_1.heading_theta
        # print(obj_1.heading_theta / np.pi * 180)
        obj_1.destroy()

        o, _ = env.reset()
        env.engine.terrain.dynamic_nodes[0].setFriction(0.)
        obj_1 = env.engine.spawn_object(Pedestrian, position=[10, 3], heading_theta=0, random_seed=1)
        start_p = obj_1.position[0]
        for _ in range(10):
            # obj_1.set_position([30,0], 10)
            # 10 s , np.pi/10 per second
            obj_1.set_velocity([1, 0])
            o, r, tm, tc, info = env.step([0, 0])
        assert abs(obj_1.position[0] - start_p) > 0.7
        # print(obj_1.heading_theta / np.pi * 180)
        obj_1.destroy()

        o_ = env.reset()
        env.engine.terrain.dynamic_nodes[0].setFriction(0.)
        obj_1 = env.engine.spawn_object(Pedestrian, position=[10, 3], heading_theta=0, random_seed=1)
        start_p = obj_1.position[1]
        for _ in range(10):
            # obj_1.set_position([30,0], 10)
            # 10 s , np.pi/10 per second
            obj_1.set_velocity([0, 1])
            o, r, tm, tc, info = env.step([0, 0])
        assert abs(obj_1.position[1] - start_p) > 0.7
        # print(obj_1.heading_theta / np.pi * 180)
        obj_1.destroy()

    finally:
        env.close()


if __name__ == "__main__":
    # test_set_angular_v_and_set_v_no_friction_pedestrian(True)
    test_set_angular_v_and_set_v_no_friction(False)
    # test_coordinates(True)
    # test_set_get_vehicle_attribute(True)



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import bpy
import copy
import decimal
import math

try:
    from collections.abc import Callable
except ImportError:
    from collections import Callable

from metadrive.third_party.kitsunetsuki.base.matrices import get_bone_matrix, quat_to_list

from . import spec


class AnimationMixin(object):
    def _make_sampler(self, path, input_id, bone):
        # transforms
        channel = self._buffer.add_channel(
            {
                'componentType': spec.TYPE_FLOAT,
                'type': 'VEC4' if path == 'rotation' else 'VEC3',
                'extras': {
                    'reference': path,
                },
            }
        )

        gltf_sampler = {
            'interpolation': 'LINEAR',
            'input': input_id,
            'output': channel['bufferView'],
            'extras': {
                'joint': bone.name,
            }
        }
        return gltf_sampler

    def make_action(self, node, armature, action):
        gltf_armature = None
        for i, gltf_node in enumerate(self._root['nodes']):
            if gltf_node['name'] == armature.name:
                gltf_armature = gltf_node
                break
        if not gltf_armature:
            gltf_armature = self.make_armature(node, armature)

        gltf_skin = None
        for i in gltf_armature['children']:
            gltf_node = self._root['nodes'][i]
            if 'skin' in gltf_node:
                gltf_skin = self._root['skins'][gltf_node['skin']]
                break

        if not gltf_skin:
            print('FAILED TO FIND GLTF SKIN')
            return

        # <-- animation
        gltf_animation = {
            'name': action.name if action else 'GLTF_ANIMATION',
            'channels': [],
            'samplers': [],
        }

        # setup bones
        gltf_channels = {}
        gltf_samplers = []
        for gltf_joint_id in gltf_skin['joints']:
            gltf_joint = self._root['nodes'][gltf_joint_id]
            bone = armature.data.bones[gltf_joint['name']]

            gltf_target = {}
            if gltf_joint_id is not None:
                gltf_target['node'] = gltf_joint_id

            # time or animation frame
            channel = self._buffer.add_channel(
                {
                    'componentType': spec.TYPE_FLOAT,
                    'type': 'SCALAR',
                    'extras': {
                        'reference': 'input',
                    },
                }
            )
            input_id = channel['bufferView']

            for path in ('rotation', 'scale', 'translation'):
                gltf_samplers.append(self._make_sampler(path, input_id, bone))

                gltf_channel = {
                    'sampler': len(gltf_samplers) - 1,
                    'target': copy.copy(gltf_target),
                    'extras': {
                        'joint': bone.name,
                    }
                }
                gltf_channel['target']['path'] = path
                gltf_channels['{}/{}'.format(bone.name, path)] = gltf_channel

        gltf_animation['channels'] = list(gltf_channels.values())
        gltf_animation['samplers'] = gltf_samplers

        # set animation data
        frame_start = bpy.context.scene.frame_start
        frame_end = bpy.context.scene.frame_end
        if action and armature.animation_data:
            armature.animation_data.action = action
            frame_start, frame_end = action.frame_range

        frame = float(frame_start)
        frame_int = None
        t = decimal.Decimal(0)
        fps = bpy.context.scene.render.fps / bpy.context.scene.render.fps_base
        dt = decimal.Decimal(1 / fps)
        t += dt
        while frame <= frame_end:
            # switch frame
            if frame_int != math.floor(frame):
                frame_int = math.floor(frame)
                bpy.context.scene.frame_current = frame_int
                bpy.context.scene.frame_set(frame_int)

            if isinstance(self._speed_scale, Callable):
                speed_scale = self._speed_scale(frame_int)
            else:
                speed_scale = self._speed_scale

            # switch subframe
            if speed_scale != 1:
                bpy.context.scene.frame_subframe = frame - frame_int

            # write bone matrices
            for gltf_joint_id in gltf_skin['joints']:
                gltf_joint = self._root['nodes'][gltf_joint_id]
                bone = armature.pose.bones[gltf_joint['name']]
                bone_matrix = self._transform(get_bone_matrix(bone, armature))

                rotation = quat_to_list(bone_matrix.to_quaternion())
                scale = list(bone_matrix.to_scale())
                translation = list(bone_matrix.to_translation())

                gltf_channel = gltf_channels['{}/{}'.format(bone.name, 'rotation')]
                gltf_sampler = gltf_samplers[gltf_channel['sampler']]
                self._buffer.write(gltf_sampler['output'], *rotation)

                gltf_channel = gltf_channels['{}/{}'.format(bone.name, 'scale')]
                gltf_sampler = gltf_samplers[gltf_channel['sampler']]
                self._buffer.write(gltf_sampler['output'], *scale)

                gltf_channel = gltf_channels['{}/{}'.format(bone.name, 'translation')]
                gltf_sampler = gltf_samplers[gltf_channel['sampler']]
                self._buffer.write(gltf_sampler['output'], *translation)

                self._buffer.write(gltf_sampler['input'], float(t))

            # advance to the next frame
            frame += speed_scale
            t += dt

        # animation -->

        self._root['animations'].append(gltf_animation)



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import bpy
import bmesh


def obj2mesh(obj, triangulate=True):
    # convert the object to a mesh, so we can read the polygons
    # https://docs.blender.org/api/blender2.8/bpy.types.Object.html?highlight=to_mesh#bpy.types.Object.to_mesh
    # mesh = obj.to_mesh(
    #     bpy.context.depsgraph, apply_modifiers=True, calc_undeformed=True)
    mesh = obj.to_mesh()

    # get a BMesh representation
    b_mesh = bmesh.new()
    b_mesh.from_mesh(mesh)

    # triangulate the mesh
    if triangulate:
        bmesh.ops.triangulate(b_mesh, faces=b_mesh.faces)

    # copy the bmesh back to the original mesh
    b_mesh.to_mesh(mesh)

    # calculate the per-vertex normals, in case blender did not do that yet.
    mesh.calc_normals()
    mesh.calc_normals_split()

    return mesh



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

# Disable the xxx has no yyy member warning, pylint seems to be
# unable to figure out the Colorama properties, and throws an error
# for each property.
# pylint: disable=no-member

from __future__ import print_function

import sys

# Load and init colorama, used to color the output
from metadrive.render_pipeline.rplibs.colorama import init as init_colorama
from metadrive.render_pipeline.rplibs.colorama import Fore, Style
init_colorama()


class RPObject(object):
    """ This is the base class for every object in the render pipeline. It
    provides the functions debug, warn, error and fatal for classes which
    inherit from this object, including the name of the class when printing out
    the message. """

    _OUTPUT_LEVEL = 0
    _OUTPUT_LEVELS = ["debug", "warning", "error", "fatal"]

    @classmethod
    def set_output_level(cls, level):
        """ Sets the output level, messages with a level below will not be
        printed. E.g. if you set the output level to "error", only error and
        fatal messages will be shown.  """
        assert level in RPObject._OUTPUT_LEVELS
        RPObject._OUTPUT_LEVEL = RPObject._OUTPUT_LEVELS.index(level)

    @staticmethod
    def global_debug(context, *args, **kwargs):
        """ This method can be used from a static context to print a debug
        message. The first argument should be the name of the object / context,
        all other arguments should be the message. """
        if RPObject._OUTPUT_LEVEL > 0:
            return
        print(
            kwargs.get("color", Fore.GREEN) + "[>] " + context.ljust(25) + " " + Style.RESET_ALL + Fore.WHITE +
            ' '.join([str(i) for i in args]), Fore.RESET + Style.RESET_ALL
        )

    @staticmethod
    def global_warn(context, *args):
        """ This method can be used from a static context to print a warning.
        The first argument should be the name of the object / context, all
        other arguments should be the message. """
        if RPObject._OUTPUT_LEVEL > 1:
            return
        print(
            Fore.YELLOW + Style.BRIGHT + "[!] " + context.ljust(25) + Fore.YELLOW + Style.BRIGHT + " " +
            ' '.join([str(i) for i in args]) + Fore.RESET + Style.RESET_ALL
        )

    @staticmethod
    def global_error(context, *args):
        """ This method can be used from a static context to print an error.
        The first argument should be the name of the object / context, all
        other arguments should be the message. """
        if RPObject._OUTPUT_LEVEL > 2:
            return
        print(
            Fore.RED + Style.BRIGHT + "\n[!!!] " + context.ljust(23) + " " + ' '.join([str(i) for i in args]) + "\n" +
            Fore.RESET + Style.RESET_ALL
        )

    def __init__(self, name=None):
        """ Initiates the RPObject with a given name. The name should be
        representative about the class. If no name is given, the classname
        is used """
        if name is None:
            name = self.__class__.__name__
        self._debug_name = name
        self._debug_color = Fore.GREEN

    def _set_debug_color(self, color, style=None):
        """ Sets the color used to output debug messages """
        self._debug_color = getattr(Fore, color.upper())
        if style:
            self._debug_color += getattr(Style, style.upper())

    @property
    def debug_name(self):
        """ Returns the name of the debug object """
        return self._debug_name

    @debug_name.setter
    def debug_name(self, name):
        """ Renames this object """
        self._debug_name = name

    def debug(self, *args):
        """ Outputs a debug message, something that is not necessarry
        but provides useful information for the developer """
        self.global_debug(self._debug_name, *args, color=self._debug_color)

    def warn(self, *args):
        """ Outputs a warning message, something that failed or does
        not work, but does not prevent the program from running """
        self.global_warn(self._debug_name, *args)

    def error(self, *args):
        """ Outputs an error message, something really serious.
        Hopefully this never get's called! """
        self.global_error(self._debug_name, *args)

    def fatal(self, *args):
        """ Outputs a fatal error message, printing out the errors and then
        calling sys.exit to terminate the program. This method should be called
        when something failed so hard that the pipeline has to exit. """
        # We have to set output level to 0 here, so we can print out errors
        RPObject._OUTPUT_LEVEL = 0
        self.error(*args)
        sys.exit(0)

    def __repr__(self):
        """ Represents this object. Subclasses can override this """
        return self._debug_name
